// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Tridiagonal solver
// Method: cyclic reduction
// Size of matrix: power of two
// GPU code: global memory
// Compile: nvcc -arch=sm_20 -O3 cyclic_reduction_global_memory.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define n 20
#define N (1<<n)          // size of matrix (power 2^n) 
#define m 8     
#define BLOCK_SIZE (1<<m) // size of cuda block (power 2^m)
#define BASETYPE float    // float or double

// ---------------------------------------------------------------------  //
void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void forward_reduction(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, int a)
{ unsigned int tid=threadIdx.x;
  unsigned int idx = a*(threadIdx.x + blockDim.x * blockIdx.x)+a-1;
  unsigned int idx1 = idx-a;
  unsigned int idx2 = idx+a;
  BASETYPE x,y;
  
  if((tid%2==1)&&(idx>a-1)&&(idx<N)) x= -A[idx]/B[idx1];
  if((tid%2==1)&&(idx<N-a))  y=-C[idx]/B[idx2];
  if((tid%2==1)&&(idx>2*a-1)&&(idx<N)) A[idx]= x*A[idx1]; 
    else if(idx==2*a-1) A[idx]=0;
  if((tid%2==1)&&(idx>a-1)&&(idx<N-a))  B[idx] += x*C[idx1]+y*A[idx2]; 
    else if(idx==N-1) B[idx] += x*C[idx1];
  if((tid%2==1)&&(idx<N-a))  C[idx]= y*C[idx2]; 
    else if(idx==N-1) C[idx]=0;
  if((tid%2==1)&&(idx>a-1)&&(idx<N-a))  D[idx] += x*D[idx1]+y*D[idx2]; 
    else if(idx==N-1) D[idx] += x*D[idx1];
}

__global__ void init_backward_substitution(BASETYPE * B, BASETYPE *D, BASETYPE *X)
{ if ((threadIdx.x==0)&&(blockIdx.x==0)) X[N-1]=D[N-1]/B[N-1]; }

__global__ void backward_substitution(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, BASETYPE* X, int a)
{ unsigned int tid=threadIdx.x;
  unsigned int idx = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int idx1 = idx-a;
  unsigned int idx2 = idx+a;
  
  if(idx==a-1) X[idx]=(D[idx]-C[idx]*X[idx2])/B[idx];
  else if((tid%2==0)&&(idx>a-1)&&(idx<N-a))
               X[idx]=(D[idx]-C[idx]*X[idx2]-A[idx]*X[idx1])/B[idx];
}

void CPU_cyclic_reduction(BASETYPE *A, BASETYPE *B, BASETYPE *C, BASETYPE *D, BASETYPE *X)
{  int i, j, k; BASETYPE *A1,*B1,*C1,*D1, a, b;

   A1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   B1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   C1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   D1=(BASETYPE *)malloc(N*sizeof(BASETYPE));

   // Aprovecho para cambiar el inicio de A
   memcpy(A1+1,A,(N-1)*sizeof(BASETYPE)); memcpy(B1,B,N*sizeof(BASETYPE));
   memcpy(C1,C,N*sizeof(BASETYPE)); memcpy(D1,D,N*sizeof(BASETYPE));
   
   for(i=0,k=1;i<n;i++,k*=2)
   {  for(j=2*k-1;j<N;j+=2*k)
      {  a=-A1[j]/B1[j-k];
         if(j!=N-1) b=-C1[j]/B1[j+k];
	 if(j!=2*k-1) A1[j]=a*A1[j-k]; else A1[j]=0;
         if(j!=N-1) B1[j] += a*C1[j-k]+b*A1[j+k]; else B1[j] += a*C1[j-k];
	 if(j!=N-1) C1[j]= b*C1[j+k]; else C1[j]=0;
	 if(j!=N-1) D1[j] += a*D1[j-k]+b*D1[j+k]; else  D1[j] += a*D1[j-k];
      }
   }   
   
   X[N-1]=D1[N-1]/B1[N-1];
   for(i=0,k=N/2;i<n;i++,k/=2)
   {  for(j=k-1;j<N;j+=2*k)
      {  if(j==k-1) X[j]=(D1[j]-C1[j]*X[j+k])/B1[j];
	 else       X[j]=(D1[j]-C1[j]*X[j+k]-A1[j]*X[j-k])/B1[j];
      }
   }
}

double matrix_mult_error(BASETYPE *A, BASETYPE *B, BASETYPE *C, BASETYPE *D, BASETYPE *X)
{  double dc, err=0; int i;
   dc=B[0]*X[0]+C[0]*X[1]; err += fabs(D[0]-dc);
   for(i=1;i<N-1;i++) { dc=A[i-1]*X[i-1]+B[i]*X[i]+C[i]*X[i+1];
      err += fabs(D[i]-dc); }
   dc=A[N-2]*X[N-2]+B[N-1]*X[N-1]; err += fabs(D[N-1]-dc);
   return(err);
}  

int main()
{  BASETYPE *A, *B, *C, *D, *X, *X1, *Ad, *Bd, *Cd, *Dd, *Xd;
   int i, j, a; double error; struct timeval start, end;

   A = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   B = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   C = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   D = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   X = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   X1= (BASETYPE *) malloc(N*sizeof(BASETYPE));

   // Input data
   for(i=0;i<N;i++) { A[i]=drand48(); B[i]=drand48(); 
                      C[i]=drand48(); D[i]=drand48(); } // Example of input data

   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ad,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Bd,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Cd,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Dd,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Xd,N*sizeof(BASETYPE));
   cudaMemcpy(Ad+1,A,(N-1)*sizeof(BASETYPE),cudaMemcpyHostToDevice); // Shift of A
   cudaMemcpy(Bd,B,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Cd,C,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Dd,D,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);

   dim3 dimBlock0(BLOCK_SIZE);
   dim3 dimGrid0(N/BLOCK_SIZE);
   
   // Computation - forward reduction
   gettimeofday(&start,NULL);
   for(j=0,a=1;j<n-m-5;j++,a*=2)
   {  dim3 dimGrid1(N/(a*BLOCK_SIZE));
      forward_reduction<<<dimGrid1,dimBlock0>>>(Ad,Bd,Cd,Dd,a); }
   for(j=n-m-5;j<n-5;j++,a*=2)
   {  dim3 dimBlock2(N/(a*32));
      dim3 dimGrid2(32);
      forward_reduction<<<dimGrid2,dimBlock2>>>(Ad,Bd,Cd,Dd,a); }
   for(j=n-5;j<n;j++,a*=2)
   {  dim3 dimBlock3(32);
      dim3 dimGrid3(1);
      forward_reduction<<<dimGrid3,dimBlock3>>>(Ad,Bd,Cd,Dd,a); }

   // Computation - backward substitution
   dim3 dimGrid1(1);
   init_backward_substitution<<<dimGrid1,dimBlock0>>>(Bd,Dd,Xd);
   for(j=0,a=N/2;j<m;j++,a/=2)
   {  dim3 dimGrid1(1);
      backward_substitution<<<dimGrid1,dimBlock0>>>(Ad,Bd,Cd,Dd,Xd,a); }
   for(j=m;j<n;j++,a/=2)
   {  dim3 dimGrid1(N/(a*BLOCK_SIZE));
      backward_substitution<<<dimGrid1,dimBlock0>>>(Ad,Bd,Cd,Dd,Xd,a); }
   cudaThreadSynchronize(); 
   gettimeofday(&end,NULL);
   printf("GPU time=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   
   // Write results
   cudaMemcpy(X1,Xd,N*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ad); cudaFree(Bd); cudaFree(Cd); cudaFree(Dd); cudaFree(Xd);
   
   // CPU test
   gettimeofday(&start,NULL);
   CPU_cyclic_reduction(A,B,C,D,X);
   gettimeofday(&end,NULL);
   printf("CPU time=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   cuda_error("Execution error");
   error=0; for(i=0;i<N;i++) error += fabs(X[i]-X1[i]);

   printf("Cyclic reduction, global memory\n");
   printf("Matrix size= %d, Floatsize=%ld\n",N,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(CPU-GPU)= %f\n",error);
   printf("Acumulative error(MULTI)= %f\n",matrix_mult_error(A,B,C,D,X1));

   free(A); free(B); free(C); free(D); free(X); free(X1); 
}
         
   
