// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Tridiagonal solver
// Method: cyclic reduction
// Size of matrix: power of two
// GPU code: split shared, merge shared, half of threads
// Compile: nvcc -arch=sm_20 -O3 cyclic_reduction_split_and_merge.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define n 20     
#define N (1<<n)          // size of matrix (power 2^n) 
#define m 8     
#define BLOCK_SIZE (1<<m) // size of cuda block (power 2^m)
#define W 0               // split-and-merge parameter
#define BASETYPE float    // float or double

// ---------------------------------------------------------------------  //
void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void forward_split_shared(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, int a)
{   __shared__ BASETYPE As[BLOCK_SIZE-W], Bs[BLOCK_SIZE-W], Cs[BLOCK_SIZE-W], Ds[BLOCK_SIZE-W];
  unsigned int tid= 2*threadIdx.x+1;
  unsigned int idx= a*((2*threadIdx.x+1) + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int tid3= threadIdx.x;
  unsigned int idx3= a*(threadIdx.x + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int tid1, tid2;
  BASETYPE x, y;

  As[tid3]=A[idx3]; Bs[tid3]=B[idx3]; Cs[tid3]=C[idx3]; Ds[tid3]=D[idx3]; 
  As[tid3+(BLOCK_SIZE-W)/2]=A[idx3+a*(BLOCK_SIZE-W)/2];
  Bs[tid3+(BLOCK_SIZE-W)/2]=B[idx3+a*(BLOCK_SIZE-W)/2];
  Cs[tid3+(BLOCK_SIZE-W)/2]=C[idx3+a*(BLOCK_SIZE-W)/2];
  Ds[tid3+(BLOCK_SIZE-W)/2]=D[idx3+a*(BLOCK_SIZE-W)/2]; 
  __syncthreads(); 
  
  // one stage
  tid1=tid-1;  tid2=tid+1;
  if(tid<BLOCK_SIZE-W-1) 
  {  x= -As[tid]/Bs[tid1]; y=-Cs[tid]/Bs[tid2];
     As[tid]= x*As[tid1]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; }
  __syncthreads(); 
    if(idx==2*a-1) As[1]=0;
  __syncthreads(); 
 
  // two stages
  tid1=tid-2;  tid2=tid+2;
  if((tid%4==3)&&(tid<BLOCK_SIZE-W-1)) 
  {  x= -As[tid]/Bs[tid1]; y=-Cs[tid]/Bs[tid2];
     As[tid]= x*As[tid1]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; }
  __syncthreads(); 
    if(idx==4*a-1) As[3]=0;
  __syncthreads(); 
 
  // three stages
  tid1=tid-4;  tid2=tid+4;
  if((tid%8==7)&&(tid<BLOCK_SIZE-W-1)) 
  {  x= -As[tid]/Bs[tid1];  y=-Cs[tid]/Bs[tid2];
     As[tid]= x*As[tid1]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; }
  __syncthreads(); 
    if(idx==8*a-1) As[7]=0;
  __syncthreads(); 
 
  // four stages
  tid1=tid-8;  tid2=tid+8;
  if((tid%16==15)&&(tid<BLOCK_SIZE-W-1)) 
  {  x= -As[tid]/Bs[tid1]; y=-Cs[tid]/Bs[tid2];
     As[tid]= x*As[tid1]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; }
  __syncthreads(); 
    if(idx==16*a-1) As[15]=0;
  __syncthreads(); 

  A[idx3]=As[tid3]; B[idx3]=Bs[tid3]; C[idx3]=Cs[tid3]; D[idx3]=Ds[tid3]; 
  A[idx3+a*(BLOCK_SIZE-W)/2]=As[tid3+(BLOCK_SIZE-W)/2];
  B[idx3+a*(BLOCK_SIZE-W)/2]=Bs[tid3+(BLOCK_SIZE-W)/2];
  C[idx3+a*(BLOCK_SIZE-W)/2]=Cs[tid3+(BLOCK_SIZE-W)/2];
  D[idx3+a*(BLOCK_SIZE-W)/2]=Ds[tid3+(BLOCK_SIZE-W)/2]; 
}

__global__ void forward_merge_shared(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, int a)
{ __shared__ BASETYPE As[18+W], Bs[18+W], Cs[18+W], Ds[18+W];
  unsigned int tid= 2*threadIdx.x;
  unsigned int idx= (a*(2*threadIdx.x + BLOCK_SIZE *(blockIdx.x+1)-9-W)+a-1)%N;
  unsigned int tid3= threadIdx.x;
  unsigned int idx3= (a*(threadIdx.x + BLOCK_SIZE *(blockIdx.x+1)-9-W)+a-1)%N;
  unsigned int tid1, tid2;
  BASETYPE x, y; 

  As[tid3]=A[idx3]; Bs[tid3]=B[idx3]; Cs[tid3]=C[idx3]; Ds[tid3]=D[idx3]; 
  As[tid3+(18+W)/2]=A[idx3+a*(18+W)/2];
  Bs[tid3+(18+W)/2]=B[idx3+a*(18+W)/2];
  Cs[tid3+(18+W)/2]=C[idx3+a*(18+W)/2];
  Ds[tid3+(18+W)/2]=D[idx3+a*(18+W)/2]; 
  __syncthreads(); 

  // one stage
  tid1=tid-1;  tid2=tid+1;
  if((tid%2==0)&&(tid>7)&&(tid<9+W)&&(idx<N)) 
  {  x= -As[tid]/Bs[tid1]; As[tid]= x*As[tid1]; }
  if((tid%2==0)&&(tid>7)&&(tid<9+W)&&(idx<N-a))  
  {  y=-Cs[tid]/Bs[tid2]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; } 
  else if(idx==N-1) { Bs[tid] += x*Cs[tid1]; Cs[tid]=0; Ds[tid] += x*Ds[tid1]; }
  __syncthreads(); 

  // two stages
  tid1=tid-2;  tid2=tid+2;
  if((tid%4==0)&&(tid>7)&&(tid<9+W)&&(idx<N)) 
  {  x= -As[tid]/Bs[tid1];  As[tid]= x*As[tid1]; } 
  if((tid%4==0)&&(tid>7)&&(tid<9+W)&&(idx<N-2*a)) 
  {  y=-Cs[tid]/Bs[tid2]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; } 
  else if(idx==N-1) { Bs[tid] += x*Cs[tid1]; Cs[tid]=0; Ds[tid] += x*Ds[tid1]; }
  __syncthreads(); 

  // three stages
  tid1=tid-4;  tid2=tid+4;
  if((tid%8==0)&&(tid>7)&&(tid<9+W)&&(idx<N)) 
  {  x= -As[tid]/Bs[tid1]; As[tid]= x*As[tid1]; }
  if((tid%8==0)&&(tid>7)&&(tid<9+W)&&(idx<N-4*a)) 
  {  y=-Cs[tid]/Bs[tid2]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; }
    else if(idx==N-1) { Bs[tid] += x*Cs[tid1]; Cs[tid]=0; Ds[tid] += x*Ds[tid1]; }
  __syncthreads(); 

  // four stages
  tid1=tid-8;  tid2=tid+8;
  if((tid%16==8)&&(tid>7)&&(tid<9+W)&&(idx<N)) 
  {  x= -As[tid]/Bs[tid1];  As[tid]= x*As[tid1]; } 
  if((tid%16==8)&&(tid>7)&&(tid<9+W)&&(idx<N-8*a))  
  {  y=-Cs[tid]/Bs[tid2]; Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
     Cs[tid]= y*Cs[tid2]; Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; }
  else if(idx==N-1) { Cs[tid]=0; Bs[tid] += x*Cs[tid1]; Ds[tid] += x*Ds[tid1]; }
  __syncthreads(); 

  A[idx3]=As[tid3]; B[idx3]=Bs[tid3]; C[idx3]=Cs[tid3]; D[idx3]=Ds[tid3]; 
  A[idx3+a*(18+W)/2]=As[tid3+(18+W)/2];
  B[idx3+a*(18+W)/2]=Bs[tid3+(18+W)/2];
  C[idx3+a*(18+W)/2]=Cs[tid3+(18+W)/2];
  D[idx3+a*(18+W)/2]=Ds[tid3+(18+W)/2]; 
}

__global__ void forward_intermediate_stages_global(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, int a)
{ unsigned int idx = a*((2*threadIdx.x+1) + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int idx1 = idx-a;
  unsigned int idx2 = idx+a;
  BASETYPE x,y;
  
  if((idx>a-1)&&(idx<N)) x= -A[idx]/B[idx1];
  if((idx<N-a))  y=-C[idx]/B[idx2];
  if((idx>2*a-1)&&(idx<N)) A[idx]= x*A[idx1]; 
    else if(idx==2*a-1) A[idx]=0;
  if((idx>a-1)&&(idx<N-a))  B[idx] += x*C[idx1]+y*A[idx2]; 
    else if(idx==N-1) B[idx] += x*C[idx1];
  if((idx<N-a))  C[idx]= y*C[idx2]; 
    else if(idx==N-1) C[idx]=0;
  if((idx>a-1)&&(idx<N-a))  D[idx] += x*D[idx1]+y*D[idx2]; 
    else if(idx==N-1) D[idx] += x*D[idx1];
}

__global__ void forward_final_stages_shared(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, int a)
{   __shared__ BASETYPE As[BLOCK_SIZE], Bs[BLOCK_SIZE], Cs[BLOCK_SIZE], Ds[BLOCK_SIZE];
  unsigned int tid= 2*threadIdx.x+1;
  unsigned int idx= a*((2*threadIdx.x+1) + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int tid3= threadIdx.x;
  unsigned int idx3= a*(threadIdx.x + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int tid1, tid2;
  BASETYPE x, y;

  As[tid3]=A[idx3]; Bs[tid3]=B[idx3]; Cs[tid3]=C[idx3]; Ds[tid3]=D[idx3]; 
  As[tid3+blockDim.x]=A[idx3+a*blockDim.x];
  Bs[tid3+blockDim.x]=B[idx3+a*blockDim.x];
  Cs[tid3+blockDim.x]=C[idx3+a*blockDim.x];
  Ds[tid3+blockDim.x]=D[idx3+a*blockDim.x]; 
 __syncthreads(); 
  
  // one stage
  tid1=tid-1;  tid2=tid+1;
  if((tid%2==1)&&(idx>a-1)&&(idx<N)) x= -As[tid]/Bs[tid1];
  if((tid%2==1)&&(idx<N-a))  y=-Cs[tid]/Bs[tid2];
  if((tid%2==1)&&(idx>2*a-1)&&(idx<N)) As[tid]= x*As[tid1]; 
    else if(idx==2*a-1) As[tid]=0;
  if((tid%2==1)&&(idx>a-1)&&(idx<N-a))  Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
    else if(idx==N-1) Bs[tid] += x*Cs[tid1];
  if((tid%2==1)&&(idx<N-a))  Cs[tid]= y*Cs[tid2]; 
    else if(idx==N-1) Cs[tid]=0;
  if((tid%2==1)&&(idx>a-1)&&(idx<N-a))  Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; 
    else if(idx==N-1) Ds[tid] += x*Ds[tid1];

  // two stages
  tid1=tid-2;  tid2=tid+2;
  if((tid%4==3)&&(idx>2*a-1)&&(idx<N)) x= -As[tid]/Bs[tid1];
  if((tid%4==3)&&(idx<N-2*a))  y=-Cs[tid]/Bs[tid2];
  if((tid%4==3)&&(idx>4*a-1)&&(idx<N)) As[tid]= x*As[tid1]; 
    else if(idx==4*a-1) As[tid]=0;
  if((tid%4==3)&&(idx>2*a-1)&&(idx<N-2*a))  Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
    else if(idx==N-1) Bs[tid] += x*Cs[tid1];
  if((tid%4==3)&&(idx<N-2*a))  Cs[tid]= y*Cs[tid2]; 
    else if(idx==N-1) Cs[tid]=0;
  if((tid%4==3)&&(idx>2*a-1)&&(idx<N-2*a))  Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; 
    else if(idx==N-1) Ds[tid] += x*Ds[tid1];

  // three stages
  tid1=tid-4;  tid2=tid+4;
  if((tid%8==7)&&(idx>4*a-1)&&(idx<N)) x= -As[tid]/Bs[tid1];
  if((tid%8==7)&&(idx<N-4*a))  y=-Cs[tid]/Bs[tid2];
  if((tid%8==7)&&(idx>8*a-1)&&(idx<N)) As[tid]= x*As[tid1]; 
    else if(idx==4*a-1) As[tid]=0;
  if((tid%8==7)&&(idx>4*a-1)&&(idx<N-4*a))  Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
    else if(idx==N-1) Bs[tid] += x*Cs[tid1];
  if((tid%8==7)&&(idx<N-4*a))  Cs[tid]= y*Cs[tid2]; 
    else if(idx==N-1) Cs[tid]=0;
  if((tid%8==7)&&(idx>4*a-1)&&(idx<N-a))  Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; 
    else if(idx==N-1) Ds[tid] += x*Ds[tid1];

  // four stages
  tid1=tid-8;  tid2=tid+8;
  if((tid%16==15)&&(idx>8*a-1)&&(idx<N)) x= -As[tid]/Bs[tid1];
  if((tid%16==15)&&(idx<N-8*a))  y=-Cs[tid]/Bs[tid2];
  if((tid%16==15)&&(idx>16*a-1)&&(idx<N)) As[tid]= x*As[tid1]; 
    else if(idx==2*a-1) As[tid]=0;
  if((tid%16==15)&&(idx>8*a-1)&&(idx<N-8*a))  Bs[tid] += x*Cs[tid1]+y*As[tid2]; 
    else if(idx==N-1) Bs[tid] += x*Cs[tid1];
  if((tid%16==15)&&(idx<N-8*a))  Cs[tid]= y*Cs[tid2]; 
    else if(idx==N-1) Cs[tid]=0;
  if((tid%16==15)&&(idx>8*a-1)&&(idx<N-8*a))  Ds[tid] += x*Ds[tid1]+y*Ds[tid2]; 
    else if(idx==N-1) Ds[tid] += x*Ds[tid1];
  __syncthreads(); 

  A[idx3]=As[tid3]; B[idx3]=Bs[tid3]; C[idx3]=Cs[tid3]; D[idx3]=Ds[tid3]; 
  A[idx3+a*blockDim.x]=As[tid3+blockDim.x];
  B[idx3+a*blockDim.x]=Bs[tid3+blockDim.x];
  C[idx3+a*blockDim.x]=Cs[tid3+blockDim.x];
  D[idx3+a*blockDim.x]=Ds[tid3+blockDim.x]; 
}

__global__ void backward_init(BASETYPE * B, BASETYPE *D, BASETYPE *X)
{ if ((threadIdx.x==0)&&(blockIdx.x==0)) X[N-1]=D[N-1]/B[N-1]; }

__global__ void backward_intermediate_stages_global(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, BASETYPE* X, int a)
{ unsigned int idx = a*(2*threadIdx.x + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int idx1 = idx-a;
  unsigned int idx2 = idx+a;
  
  if(idx==a-1) X[idx]=(D[idx]-C[idx]*X[idx2])/B[idx];
  else if((idx>a-1)&&(idx<N-a))
               X[idx]=(D[idx]-C[idx]*X[idx2]-A[idx]*X[idx1])/B[idx];
}

__global__ void backward_initial_stages_shared(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, BASETYPE *X, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE], Bs[BLOCK_SIZE], Cs[BLOCK_SIZE], 
                   Ds[BLOCK_SIZE], Xs[BLOCK_SIZE];
  unsigned int tid = 2*threadIdx.x+1;
  unsigned int idx = a*((2*threadIdx.x+1) + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int tid3 = threadIdx.x;
  unsigned int idx3 = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int tid1, tid2; 
  
  As[tid3]=A[idx3]; Bs[tid3]=B[idx3]; Cs[tid3]=C[idx3]; Ds[tid3]=D[idx3]; Xs[tid3]=X[idx3];
  As[tid3+blockDim.x]=A[idx3+a*blockDim.x];
  Bs[tid3+blockDim.x]=B[idx3+a*blockDim.x];
  Cs[tid3+blockDim.x]=C[idx3+a*blockDim.x];
  Ds[tid3+blockDim.x]=D[idx3+a*blockDim.x]; 
  Xs[tid3+blockDim.x]=X[idx3+a*blockDim.x]; 
  __syncthreads(); 

  // one stage
  tid1=tid-8;  tid2=tid+8;
  if((tid%16==7)&&(tid<2*blockDim.x-1)&&(idx>8)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(idx==7*a) Xs[7]=(Ds[7]-Cs[7]*Xs[15])/Bs[7];
  __syncthreads();

  // two stages
  tid1=tid-4;  tid2=tid+4;
  if((tid%8==3)&&(tid<2*blockDim.x-1)&&(idx>4)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(idx==3*a) Xs[3]=(Ds[3]-Cs[3]*Xs[7])/Bs[3];
  __syncthreads();

  // three stages
  tid1=tid-2;  tid2=tid+2;
  if((tid%4==1)&&(tid<2*blockDim.x-1)&&(idx>2)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(idx==1*a) Xs[1]=(Ds[1]-Cs[1]*Xs[3])/Bs[1];
  __syncthreads();

  // four stages
  tid = 2*threadIdx.x;
  idx = a*((2*threadIdx.x) + BLOCK_SIZE * blockIdx.x)+a-1;
  tid1=tid-1;  tid2=tid+1;
  if((tid<2*blockDim.x-1)&&(idx>1)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(idx==0) Xs[0]=(Ds[0]-Cs[0]*Xs[1])/Bs[0];
  __syncthreads();
    
  X[idx3]=Xs[tid3]; X[idx3+a*blockDim.x]=Xs[tid3+blockDim.x];
}

__global__ void backward_final_stages_shared(BASETYPE* A, BASETYPE* B, BASETYPE* C, BASETYPE* D, BASETYPE *X, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE], Bs[BLOCK_SIZE], Cs[BLOCK_SIZE], 
                   Ds[BLOCK_SIZE], Xs[BLOCK_SIZE], XS;
  unsigned int tid = 2*threadIdx.x+1;
  unsigned int tid3 = threadIdx.x;
  unsigned int idx3 = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x)+a-1;
  unsigned int tid1, tid2; 
  BASETYPE XR=0;
  
  As[tid3]=A[idx3]; Bs[tid3]=B[idx3]; Cs[tid3]=C[idx3]; Ds[tid3]=D[idx3]; Xs[tid3]=X[idx3];
  As[tid3+blockDim.x]=A[idx3+a*blockDim.x];
  Bs[tid3+blockDim.x]=B[idx3+a*blockDim.x];
  Cs[tid3+blockDim.x]=C[idx3+a*blockDim.x];
  Ds[tid3+blockDim.x]=D[idx3+a*blockDim.x]; 
  Xs[tid3+blockDim.x]=X[idx3+a*blockDim.x]; 
  if(tid3==0) XS=X[(idx3-a)%N];
  __syncthreads();
  if(idx3!=a-1) XR=XS;
  __syncthreads();
 
  // one stage
  tid1=tid-8;  tid2=tid+8;
  if((tid%16==7)&&(tid<2*blockDim.x-1)&&(tid>7)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(tid==7) Xs[7]=(Ds[7]-Cs[7]*Xs[15]-As[7]*XR)/Bs[7];
  __syncthreads();

   // two stages
 tid1=tid-4;  tid2=tid+4;
  if((tid%8==3)&&(tid<2*blockDim.x-1)&&(tid>3)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(tid==3) Xs[3]=(Ds[3]-Cs[3]*Xs[7]-As[3]*XR)/Bs[3];
  __syncthreads();

  // three stages
  tid1=tid-2;  tid2=tid+2;
  if((tid%4==1)&&(tid<2*blockDim.x-1)&&(tid>1)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(tid==1) Xs[1]=(Ds[1]-Cs[1]*Xs[3]-As[1]*XR)/Bs[1];
  __syncthreads();

  // four stages
  tid = 2*threadIdx.x;
  tid1=tid-1;  tid2=tid+1;
  if((tid<2*blockDim.x-1)&&(tid>1)) 
     Xs[tid]=(Ds[tid]-Cs[tid]*Xs[tid2]-As[tid]*Xs[tid1])/Bs[tid];
  else if(tid==0) Xs[0]=(Ds[0]-Cs[0]*Xs[1]-As[0]*XR)/Bs[0];
  __syncthreads();
    
  X[idx3]=Xs[tid3]; X[idx3+a*blockDim.x]=Xs[tid3+blockDim.x];
}

void CPU_cyclic_reduction(BASETYPE *A, BASETYPE *B, BASETYPE *C, BASETYPE *D, BASETYPE *X)
{  int i, j, k; BASETYPE *A1,*B1,*C1,*D1, a, b;
   A1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   B1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   C1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   D1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   // Aprovecho para cambiar el inicio de A
   memcpy(A1+1,A,(N-1)*sizeof(BASETYPE)); memcpy(B1,B,N*sizeof(BASETYPE));
   memcpy(C1,C,N*sizeof(BASETYPE)); memcpy(D1,D,N*sizeof(BASETYPE));
   for(i=0,k=1;i<n;i++,k*=2)
   {  for(j=2*k-1;j<N;j+=2*k)
      {  a=-A1[j]/B1[j-k];
         if(j!=N-1) b=-C1[j]/B1[j+k];
	 if(j!=2*k-1) A1[j]=a*A1[j-k]; else A1[j]=0;
         if(j!=N-1) B1[j] += a*C1[j-k]+b*A1[j+k]; else B1[j] += a*C1[j-k];
	 if(j!=N-1) C1[j]= b*C1[j+k]; else C1[j]=0;
	 if(j!=N-1) D1[j] += a*D1[j-k]+b*D1[j+k]; else  D1[j] += a*D1[j-k]; } }   
   X[N-1]=D1[N-1]/B1[N-1];
   for(i=0,k=N/2;i<n;i++,k/=2)
   {  for(j=k-1;j<N;j+=2*k)
      {  if(j==k-1) X[j]=(D1[j]-C1[j]*X[j+k])/B1[j];
	 else       X[j]=(D1[j]-C1[j]*X[j+k]-A1[j]*X[j-k])/B1[j]; } }
}

double matrix_mult_error(BASETYPE *A, BASETYPE *B, BASETYPE *C, BASETYPE *D, BASETYPE *X)
{  double dc, err=0; int i;
   dc=B[0]*X[0]+C[0]*X[1]; err += fabs(D[0]-dc);
   for(i=1;i<N-1;i++) { dc=A[i-1]*X[i-1]+B[i]*X[i]+C[i]*X[i+1];
      err += fabs(D[i]-dc); }
   dc=A[N-2]*X[N-2]+B[N-1]*X[N-1]; err += fabs(D[N-1]-dc);
   return(err);
}  

int main()
{  BASETYPE *A,*B,*C,*D,*X,*X1,*Ai,*Bi,*Ci,*Di,*Xd;
   int i, j, a; double error; struct timeval start, end;

   A = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   B = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   C = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   D = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   X = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   X1= (BASETYPE *) malloc(N*sizeof(BASETYPE));
   if((A==NULL)||(B==NULL)||(C==NULL)||(D==NULL)||(X==NULL)||(X1==NULL))
   {  fprintf(stderr,"Memory error\n"); exit(-1); }

   // Input data
   for(i=0;i<N;i++) { A[i]=drand48(); B[i]=drand48(); 
                      C[i]=drand48(); D[i]=drand48(); } // Example of input data

   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ai,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Bi,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Ci,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Di,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Xd,N*sizeof(BASETYPE));
   cudaMemcpy(Ai+1,A,(N-1)*sizeof(BASETYPE),cudaMemcpyHostToDevice);  // Shift of A
   cudaMemcpy(Bi,B,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Ci,C,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Di,D,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);

   dim3 dimBlock0(BLOCK_SIZE);   
   dim3 dimGrid0(N/BLOCK_SIZE);
   dim3 dimGrid1(1); 
   dim3 dimBlock2(BLOCK_SIZE/2);
   dim3 dimBlock4((BLOCK_SIZE-W)/2); 
   dim3 dimBlock5((18+W)/2);

   // Computation - forward reduction
   gettimeofday(&start,NULL);
   for(j=0,a=1;j+4<=n-m;j+=4,a=a*16)
   {  dim3 dimGridA(N/(a*BLOCK_SIZE)); // first 4*[(n-m)/4] stages 
      forward_split_shared<<<dimGridA,dimBlock4>>>(Ai,Bi,Ci,Di,a);
      forward_merge_shared<<<dimGridA,dimBlock5>>>(Ai,Bi,Ci,Di,a); }
   for(j=((n-m)/4)*4;j<n-m;j++,a*=2) // until completing n-m stages 
   {  forward_intermediate_stages_global<<<dimGrid0,dimBlock2>>>(Ai,Bi,Ci,Di,a); }
   for(j=n-m;j<n-(m/4)*4;j++,a*=2)   // next [m/4]*4 stages
   {  dim3 dimGrid1(1);
      forward_intermediate_stages_global<<<dimGrid1,dimBlock2>>>(Ai,Bi,Ci,Di,a); }
   for(j=n-(m/4)*4;j<n;j+=4,a*=16)   // remainder stages
   {  dim3 dimBlockA(1<<n-j-1);
      forward_final_stages_shared<<<dimGrid1,dimBlockA>>>(Ai,Bi,Ci,Di,a);  }

   // Computation - backward substitution
   backward_init<<<dimGrid1,dimBlock0>>>(Bi,Di,Xd);
   for(j=0,a=N/2;j<(m/4)*4;j+=4,a/=16)  // first [m/4]*4 stages
   {  dim3 dimBlockA(8*N/(2*a));
      backward_initial_stages_shared<<<dimGrid1,dimBlockA>>>(Ai,Bi,Ci,Di,Xd,a/8); }
   for(j=(m/4)*4;j<m;j++,a/=2)          // until completing m stages
   {  dim3 dimGrid1(1);
      backward_intermediate_stages_global<<<dimGrid1,dimBlock2>>>(Ai,Bi,Ci,Di,Xd,a); }
   for(j=m;j<n-((n-m)/4)*4;j++,a/=2)    // stages m -> n-[n-m]/4
   {  dim3 dimGridA(N/(a*BLOCK_SIZE));
      backward_intermediate_stages_global<<<dimGridA,dimBlock2>>>(Ai,Bi,Ci,Di,Xd,a); }
   for(j=n-((n-m)/4)*4;j<n;j+=4,a/=16)  // last [(n-m)/4]*4 stages
   {  dim3 dimGridA(8*N/(a*BLOCK_SIZE));
      backward_final_stages_shared<<<dimGridA,dimBlock2>>>(Ai,Bi,Ci,Di,Xd,a/8); }
   
   cudaThreadSynchronize(); 
   gettimeofday(&end,NULL);
   printf("GPU time=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
     
   // Write results
   cudaMemcpy(X1,Xd,N*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ai); cudaFree(Bi); cudaFree(Ci); cudaFree(Di); cudaFree(Xd);
   cuda_error("\n\nCUDA ERROR!!!\n\n");

   // CPU test
   gettimeofday(&start,NULL);
   CPU_cyclic_reduction(A,B,C,D,X);
   gettimeofday(&end,NULL);
   printf("CPU time=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   cuda_error("Execution error");
   error=0; for(i=0;i<N;i++) error += fabs(X[i]-X1[i]);

   printf("Cyclic reduction, split shared, merge shared, half threads\n");
   printf("Matrix size= %d, Floatsize=%ld\n",N,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(CPU-GPU)= %f\n",error);
   printf("Acumulative error(MULTI)= %f\n",matrix_mult_error(A,B,C,D,X1));
   free(A); free(B); free(C); free(D); free(X); free(X1);
}
         
   
