// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Tridiagonal solver
// Method: recursive doubling
// Size of matrix: power of two
// GPU code: split shared, merge global
// Compile: nvcc -arch=sm_20 -O3 recursive_dob_split_and_merge.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define n 20     
#define N (1<<n)          // size of matrix (power 2^n) 
#define m 8     
#define BLOCK_SIZE (1<<m) // size of cuda block (power 2^m)
#define W 0               // split-and-merge parameter
#define BASETYPE double    // float or double

// ---------------------------------------------------------------------  //
void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void merge_global(BASETYPE* A1, BASETYPE* B1, BASETYPE* C1, BASETYPE* D1, 
       BASETYPE* A2, BASETYPE* B2, BASETYPE* C2, BASETYPE* D2, int j, int N4, int N3)
{ unsigned int idx, idx1, idx2, i;  BASETYPE a,b;
  unsigned int tid= threadIdx.x;
  if(tid<blockDim.x/2) idx = threadIdx.x + BLOCK_SIZE * blockIdx.x;
  else      idx = threadIdx.x + BLOCK_SIZE * (blockIdx.x+1)-blockDim.x;
 
  idx1=idx-N4; idx2=idx+N4; i=idx>>j;
  if((tid<=2*N4-2)||(tid>=blockDim.x-(2*N4-1))) 
  {  if(i>0)    a=-A1[idx]/B1[idx1];
     if(i<N3-1) b=-C1[idx]/B1[idx2];
     if((i>0)&&(i<N3-1)) { B2[idx]=B1[idx]+a*C1[idx1]+b*A1[idx2];
                           D2[idx]=D1[idx]+a*D1[idx1]+b*D1[idx2]; }
     else if(i==0) { B2[idx]=B1[idx]+b*A1[idx2]; 
	             D2[idx]=D1[idx]+b*D1[idx2]; }
     else if(i==(N3-1)) { B2[idx]=B1[idx]+a*C1[idx1]; 
	                  D2[idx]=D1[idx]+a*D1[idx1]; }
     if(i>1)    A2[idx]=a*A1[idx1]; else A2[idx]=0;
     if(i<N3-2) C2[idx]=b*C1[idx2]; else C2[idx]=0; }
}

__global__ void split_shared(BASETYPE* A1, BASETYPE* B1, BASETYPE* C1, BASETYPE* D1, 
      BASETYPE* A2, BASETYPE* B2, BASETYPE* C2, BASETYPE* D2, int j, int N4, int N3)
{ __shared__ BASETYPE As1[BLOCK_SIZE], Bs1[BLOCK_SIZE], Cs1[BLOCK_SIZE], Ds1[BLOCK_SIZE],
      As2[BLOCK_SIZE], Bs2[BLOCK_SIZE], Cs2[BLOCK_SIZE], Ds2[BLOCK_SIZE];
  unsigned int tid= threadIdx.x;
  unsigned int idx  = threadIdx.x + BLOCK_SIZE * blockIdx.x;
  unsigned int tid1, tid2, i; BASETYPE a,b;

  As1[tid]=A1[idx]; Bs1[tid]=B1[idx]; Cs1[tid]=C1[idx]; Ds1[tid]=D1[idx]; 
  __syncthreads(); 

  // one stage
  tid1 = tid-N4; tid2 = tid+N4; i=idx>>j;
  if((tid>=1)&&(tid<=BLOCK_SIZE-2)) {
  if(i>0)    a=-As1[tid]/Bs1[tid1];
  if(i<N3-1) b=-Cs1[tid]/Bs1[tid2];
  if((i>0)&&(i<N3-1)) { Bs2[tid]=Bs1[tid]+a*Cs1[tid1]+b*As1[tid2];
                        Ds2[tid]=Ds1[tid]+a*Ds1[tid1]+b*Ds1[tid2]; }
  else if(i==0) { Bs2[tid]=Bs1[tid]+b*As1[tid2]; 
	          Ds2[tid]=Ds1[tid]+b*Ds1[tid2]; }
  else if(i==(N3-1)) { Bs2[tid]=Bs1[tid]+a*Cs1[tid1]; 
	               Ds2[tid]=Ds1[tid]+a*Ds1[tid1]; }
  if(i>1)    As2[tid]=a*As1[tid1]; else As2[tid]=0;
  if(i<N3-2) Cs2[tid]=b*Cs1[tid2]; else Cs2[tid]=0; }
  __syncthreads(); 

  // two stages
  j++; N4=2*N4; N3=N3/2; tid1 = tid-N4; tid2 = tid+N4; i=idx>>j;
  if((tid>=3)&&(tid<=BLOCK_SIZE-4)) {
  if(i>0)    a=-As2[tid]/Bs2[tid1];
  if(i<N3-1) b=-Cs2[tid]/Bs2[tid2];
  if((i>0)&&(i<N3-1)) { Bs1[tid]=Bs2[tid]+a*Cs2[tid1]+b*As2[tid2];
                        Ds1[tid]=Ds2[tid]+a*Ds2[tid1]+b*Ds2[tid2]; }
  else if(i==0) { Bs1[tid]=Bs2[tid]+b*As2[tid2]; 
	          Ds1[tid]=Ds2[tid]+b*Ds2[tid2]; }
  else if(i==(N3-1)) { Bs1[tid]=Bs2[tid]+a*Cs2[tid1]; 
	               Ds1[tid]=Ds2[tid]+a*Ds2[tid1]; }
  if(i>1)    As1[tid]=a*As2[tid1]; else As1[tid]=0;
  if(i<N3-2) Cs1[tid]=b*Cs2[tid2]; else Cs1[tid]=0; }
  __syncthreads(); 

  // three stages
  j++; N4=2*N4; N3=N3/2; tid1=tid-N4; tid2=tid+N4; i=idx>>j;
  tid1 = tid-N4; tid2 = tid+N4; i=idx>>j;
  if((tid>=7)&&(tid<=BLOCK_SIZE-8)) {
  if(i>0)    a=-As1[tid]/Bs1[tid1];
  if(i<N3-1) b=-Cs1[tid]/Bs1[tid2];
  if((i>0)&&(i<N3-1)) { Bs2[tid]=Bs1[tid]+a*Cs1[tid1]+b*As1[tid2];
                        Ds2[tid]=Ds1[tid]+a*Ds1[tid1]+b*Ds1[tid2]; }
  else if(i==0) { Bs2[tid]=Bs1[tid]+b*As1[tid2]; 
	          Ds2[tid]=Ds1[tid]+b*Ds1[tid2]; }
  else if(i==(N3-1)) { Bs2[tid]=Bs1[tid]+a*Cs1[tid1]; 
	               Ds2[tid]=Ds1[tid]+a*Ds1[tid1]; }
  if(i>1)    As2[tid]=a*As1[tid1]; else As2[tid]=0;
  if(i<N3-2) Cs2[tid]=b*Cs1[tid2]; else Cs2[tid]=0; }
  __syncthreads(); 

  // four stages
  j++; N4=2*N4; N3=N3/2; tid1=tid-N4; tid2=tid+N4; i=idx>>j;
  if((tid>=15)&&(tid<=BLOCK_SIZE-16)) {
  if(i>0)    a=-As2[tid]/Bs2[tid1];
  if(i<N3-1) b=-Cs2[tid]/Bs2[tid2];
  if((i>0)&&(i<N3-1)) { Bs1[tid]=Bs2[tid]+a*Cs2[tid1]+b*As2[tid2];
                        Ds1[tid]=Ds2[tid]+a*Ds2[tid1]+b*Ds2[tid2]; }
  else if(i==0) { Bs1[tid]=Bs2[tid]+b*As2[tid2]; 
	          Ds1[tid]=Ds2[tid]+b*Ds2[tid2]; }
  else if(i==(N3-1)) { Bs1[tid]=Bs2[tid]+a*Cs2[tid1]; 
	               Ds1[tid]=Ds2[tid]+a*Ds2[tid1]; }
  if(i>1)    As1[tid]=a*As2[tid1]; else As1[tid]=0;
  if(i<N3-2) Cs1[tid]=b*Cs2[tid2]; else Cs1[tid]=0; }
  __syncthreads(); 

  if((tid<=22)||(tid>=BLOCK_SIZE-23))
  { A2[idx]=As2[tid]; B2[idx]=Bs2[tid]; C2[idx]=Cs2[tid]; D2[idx]=Ds2[tid]; }

  A1[idx]=As1[tid]; B1[idx]=Bs1[tid]; C1[idx]=Cs1[tid]; D1[idx]=Ds1[tid];
}

__global__ void step_global_memory(BASETYPE* A1, BASETYPE* B1, BASETYPE* C1, BASETYPE* D1, 
                        BASETYPE* A2, BASETYPE* B2, BASETYPE* C2, BASETYPE* D2, 
			int j, int N4, int N3)
{ unsigned int idx = threadIdx.x + BLOCK_SIZE * blockIdx.x;
  unsigned int idx1 = idx-N4;
  unsigned int idx2 = idx+N4;
  unsigned int i=idx>>j;
  BASETYPE a,b;
  
  if(i>0)    a=-A1[idx]/B1[idx1];
  if(i<N3-1) b=-C1[idx]/B1[idx2];
  if((i>0)&&(i<N3-1)) { B2[idx]=B1[idx]+a*C1[idx1]+b*A1[idx2];
          D2[idx]=D1[idx]+a*D1[idx1]+b*D1[idx2]; }
  else if(i==0) { B2[idx]=B1[idx]+b*A1[idx2]; 
	          D2[idx]=D1[idx]+b*D1[idx2]; }
  else if(i==(N3-1)) { B2[idx]=B1[idx]+a*C1[idx1]; 
	        D2[idx]=D1[idx]+a*D1[idx1]; }
  if(i>1)    A2[idx]=a*A1[idx1]; else A2[idx]=0;
  if(i<N3-2) C2[idx]=b*C1[idx2]; else C2[idx]=0; 
}

__global__ void step_global_memory_final(BASETYPE* B, BASETYPE *D, BASETYPE *X)
{ unsigned int idx = threadIdx.x + BLOCK_SIZE * blockIdx.x;
  X[idx]=D[idx]/B[idx];
}

void CPU_recursive_doubling(BASETYPE *A, BASETYPE *B, BASETYPE *C, BASETYPE *D, BASETYPE *X)
{  int i, j, s, logN, N3, N4; 
   BASETYPE *A1,*B1,*C1,*D1, *A2,*B2,*C2,*D2, *tmp, a, b;
   A1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   B1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   C1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   D1=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   A2=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   B2=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   C2=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   D2=(BASETYPE *)malloc(N*sizeof(BASETYPE));
   memcpy(A1+1,A,(N-1)*sizeof(BASETYPE)); memcpy(B1,B,N*sizeof(BASETYPE));
   memcpy(C1,C,N*sizeof(BASETYPE)); memcpy(D1,D,N*sizeof(BASETYPE));

   logN=0; for(i=1;i<N;i=2*i) logN++;
   N3=N; N4=1;
   for(j=0;j<logN;j++)
   {  for(s=0;s<N;s++) 
      {  i=s>>j;
         if(i>0)    a=-A1[s]/B1[s-N4];
         if(i<N3-1) b=-C1[s]/B1[s+N4];
         if((i>0)&&(i<N3-1)) { B2[s]=B1[s]+a*C1[s-N4]+b*A1[s+N4];
                               D2[s]=D1[s]+a*D1[s-N4]+b*D1[s+N4]; }
         else if(i==0) { B2[s]=B1[s]+b*A1[s+N4]; 
	                 D2[s]=D1[s]+b*D1[s+N4]; }
         else if(i==(N3-1)) { B2[s]=B1[s]+a*C1[s-N4]; 
	                      D2[s]=D1[s]+a*D1[s-N4]; }
         if(i>1)    A2[s]=a*A1[s-N4]; else A2[s]=0;
         if(i<N3-2) C2[s]=b*C1[s+N4]; else C2[s]=0; 
      }
      N3=N3/2; N4=2*N4; 
      tmp=A1; A1=A2; A2=tmp; tmp=B1; B1=B2; B2=tmp; 
      tmp=C1; C1=C2; C2=tmp; tmp=D1; D1=D2; D2=tmp;
   }
   for(i=0;i<N;i++) X[i]=D1[i]/B1[i];
   free(A1); free(B1); free(C1); free(D1);
   free(A2); free(B2); free(C2); free(D2);
}

double matrix_mult_error(BASETYPE *A, BASETYPE *B, BASETYPE *C, BASETYPE *D, BASETYPE *X)
{  double dc, err=0; int i;
   dc=B[0]*X[0]+C[0]*X[1]; err += fabs(D[0]-dc);
   for(i=1;i<N-1;i++) { dc=A[i-1]*X[i-1]+B[i]*X[i]+C[i]*X[i+1];
      err += fabs(D[i]-dc); }
   dc=A[N-2]*X[N-2]+B[N-1]*X[N-1]; err += fabs(D[N-1]-dc);
   return(err);
}  

main()
{  BASETYPE *A,*B,*C,*D,*X,*X1,*Ai,*Bi,*Ci,*Di,*Xd;
   BASETYPE *A2, *B2, *C2, *D2;
   int i, j, a, b; double error; struct timeval start, end;

   A = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   B = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   C = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   D = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   X = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   X1= (BASETYPE *) malloc(N*sizeof(BASETYPE));
   if((A==NULL)||(B==NULL)||(C==NULL)||(D==NULL)||(X==NULL)||(X1==NULL))
   {  fprintf(stderr,"Memory error\n"); exit(-1); }

   // Input data
   for(i=0;i<N;i++) { A[i]=drand48(); B[i]=drand48(); 
                      C[i]=drand48(); D[i]=drand48(); } // Example of input data
   // for(i=0;i<N;i++) { A[i]=i+1; B[i]=5*(i+N); C[i]=i+2*N; D[i]=i+3*N; }    

   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ai,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Bi,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Ci,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Di,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Xd,N*sizeof(BASETYPE));
   cudaMalloc((void**)&A2,N*sizeof(BASETYPE));
   cudaMalloc((void**)&B2,N*sizeof(BASETYPE));
   cudaMalloc((void**)&C2,N*sizeof(BASETYPE));
   cudaMalloc((void**)&D2,N*sizeof(BASETYPE));
   cudaMemcpy(Ai+1,A,(N-1)*sizeof(BASETYPE),cudaMemcpyHostToDevice); // Shift of A
   cudaMemcpy(Bi,B,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Ci,C,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Di,D,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);

   dim3 dimBlock0(BLOCK_SIZE);
   dim3 dimGrid0(N/BLOCK_SIZE);
   dim3 dimBlock4(4);
   dim3 dimBlock10(10);
   dim3 dimBlock22(22);
   dim3 dimBlock46(46);

   // Computation 
   gettimeofday(&start,NULL);
   // first stages with split-and-merge
   split_shared<<<dimGrid0,dimBlock0>>>(Ai,Bi,Ci,Di,A2,B2,C2,D2,0,1,N);
   merge_global<<<dimGrid0,dimBlock4>>> (Ai,Bi,Ci,Di,A2,B2,C2,D2,0,1,N);
   merge_global<<<dimGrid0,dimBlock10>>>(A2,B2,C2,D2,Ai,Bi,Ci,Di,1,2,N/2);
   merge_global<<<dimGrid0,dimBlock22>>>(Ai,Bi,Ci,Di,A2,B2,C2,D2,2,4,N/4);
   merge_global<<<dimGrid0,dimBlock46>>>(A2,B2,C2,D2,Ai,Bi,Ci,Di,3,8,N/8);
   // remainder stages in global memory
   for(j=4,a=16,b=N/16;j<n;j+=2,a*=4,b=b/4)
   {  dim3 dimGrid1(N/(a*BLOCK_SIZE));
      step_global_memory<<<dimGrid0,dimBlock0>>>(Ai,Bi,Ci,Di,A2,B2,C2,D2,j,a,b); 
      step_global_memory<<<dimGrid0,dimBlock0>>>(A2,B2,C2,D2,Ai,Bi,Ci,Di,j+1,2*a,b/2);  }
   step_global_memory_final<<<dimGrid0,dimBlock0>>>(Bi,Di,Xd);

   cudaThreadSynchronize(); 
   cuda_error("Execution error");
   gettimeofday(&end,NULL);
   printf("GPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   
   // Write results
   cudaMemcpy(X1,Xd,N*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ai);cudaFree(Bi);cudaFree(Ci);cudaFree(Di);cudaFree(Xd);
   cudaFree(A2);cudaFree(B2);cudaFree(C2);cudaFree(D2);
   cuda_error("\n\nOJO\nhubo errores\n\n");
   
   // CPU test
   gettimeofday(&start,NULL);
   CPU_recursive_doubling(A,B,C,D,X);
   gettimeofday(&end,NULL);
   printf("CPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   error=0; for(i=0;i<N;i++) error += abs(X[i]-X1[i]);

   printf("Recursive doubling, split shared, merge global\n");
   printf("Matrix size= %d, Floatsize=%ld\n",N,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(CPU-GPU)= %f\n",error);
   printf("Acumulative error(MULTI)= %f\n",matrix_mult_error(A,B,C,D,X1));
   free(A); free(B); free(C); free(D); free(X); free(X1);
}
         
   
