// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Wavelet: 2-D Cohen-Daubechies-Feauveau 9-7 
// Method: filter-bank
// Size of matrix: power of two
// Data extension: cyclical 
// Ordering: even and odd coefficients interleaved 
// GPU code: split-and-merge, blocks 16x16 (horizontal) and 16x16 (vertical)
// Compile: nvcc -arch=sm_20 -O3 wavelet2D_97FBS_split_and_merge.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define N (1<<11)
#define N2 (N*N)        // Size of the matrix: power of two
#define STAGES 4        // Number of stages of the wavelet
#define BASETYPE double // float or double
#define BLOCK_SIZE1 16
#define BLOCK_SIZE (BLOCK_SIZE1*BLOCK_SIZE1)

#define h0 ( 0.0378284555)
#define h1 (-0.0238494650)
#define h2 (-0.1106244044)
#define h3 ( 0.3774028556)
#define h4 ( 0.8526986790)
#define h5 ( 0.3774028556)
#define h6 (-0.1106244044)
#define h7 (-0.0238494650)
#define h8 ( 0.0378284555)
#define g0 (-0.0645388826)
#define g1 ( 0.0406894176)
#define g2 ( 0.418092273)
#define g3 (-0.7884856164)
#define g4 ( 0.4180922732)
#define g5 ( 0.0406894176)
#define g6 (-0.0645388826)

void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void split_rows(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE], Bs[BLOCK_SIZE]; 
  unsigned int tid1 = threadIdx.x;
  unsigned int tid2 = threadIdx.y;
  unsigned int idx1 = a*(threadIdx.x + BLOCK_SIZE1 * blockIdx.x);
  unsigned int idx2 = a*(threadIdx.y + BLOCK_SIZE1 * blockIdx.y);
  unsigned int tid  = tid1+tid2*BLOCK_SIZE1;
  unsigned int idx  = idx1+idx2*N;
  
  As[tid]=Ai[idx];
  __syncthreads();

  if((tid1%2==0)&&(tid1<BLOCK_SIZE1-8)) 
  {  Bs[tid]  = h0*As[tid]+h1*As[tid+1]+h2*As[tid+2]+h3*As[tid+3]+h4*As[tid+4]+h5*As[tid+5]
               +h6*As[tid+6]+h7*As[tid+7]+h8*As[tid+8];
     Bs[tid+1]= g0*As[tid]+g1*As[tid+1]+g2*As[tid+2]+g3*As[tid+3]+g4*As[tid+4]+g5*As[tid+5]
               +g6*As[tid+6]; }	  
  __syncthreads(); 

  Ad[idx]= Bs[tid];
}

__global__ void split_columns(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE], Bs[BLOCK_SIZE]; 
  unsigned int tid1 = threadIdx.x;
  unsigned int tid2 = threadIdx.y;
  unsigned int idx1 = a*(threadIdx.x + BLOCK_SIZE1 * blockIdx.x);
  unsigned int idx2 = a*(threadIdx.y + BLOCK_SIZE1 * blockIdx.y);
  unsigned int tid  = tid1+tid2*BLOCK_SIZE1;
  unsigned int idx  = idx1+idx2*N;
  
  As[tid]=Ai[idx];
  __syncthreads();

  if((tid2%2==0)&&(tid2<BLOCK_SIZE1-8))
  {  Bs[tid]   = h0*As[tid]+h1*As[tid+16]+h2*As[tid+32]+h3*As[tid+48]+h4*As[tid+64]+h5*As[tid+80]
               +h6*As[tid+96]+h7*As[tid+112]+h8*As[tid+128];
     Bs[tid+16]= g0*As[tid]+g1*As[tid+16]+g2*As[tid+32]+g3*As[tid+48]+g4*As[tid+64]+g5*As[tid+80]
               +g6*As[tid+96]; }	  
  __syncthreads(); 

  Ad[idx]= Bs[tid];
}

__global__ void merge_rows(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[16*BLOCK_SIZE1], Bs[16*BLOCK_SIZE1]; 
  unsigned int tid1 = threadIdx.x;
  unsigned int tid2 = threadIdx.y;
  unsigned int idx1 = (a*(BLOCK_SIZE1 * (blockIdx.x+1) + threadIdx.x-8))%N;
  unsigned int idx2 = a*(threadIdx.y + BLOCK_SIZE1 * blockIdx.y);
  unsigned int tid  = tid1+tid2*16;
  unsigned int idx  = idx1+idx2*N;
  
  unsigned int tida  = (tid1+8)+tid2*16;
  unsigned int idxa  = (idx1+8*a)%N+idx2*N;

  As[tid]=Ai[idx]; As[tida]=Ai[idxa]; 
  __syncthreads();
  if(tid1%2==0)
  {  Bs[tid]  = h0*As[tid]+h1*As[tid+1]+h2*As[tid+2]+h3*As[tid+3]+h4*As[tid+4]+h5*As[tid+5]
               +h6*As[tid+6]+h7*As[tid+7]+h8*As[tid+8];
     Bs[tid+1]= g0*As[tid]+g1*As[tid+1]+g2*As[tid+2]+g3*As[tid+3]+g4*As[tid+4]+g5*As[tid+5]
               +g6*As[tid+6]; }
  __syncthreads();
  Ad[idx] = Bs[tid];
}

__global__ void merge_columns(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[16*BLOCK_SIZE1], Bs[16*BLOCK_SIZE1]; 
  unsigned int tid1 = threadIdx.x;
  unsigned int tid2 = threadIdx.y;
  unsigned int idx1 = a*(threadIdx.x + BLOCK_SIZE1 * blockIdx.x);
  unsigned int idx2 = (a*(BLOCK_SIZE1 * (blockIdx.y+1) + threadIdx.y-8))%N;
  unsigned int tid  = tid1+tid2*BLOCK_SIZE1;
  unsigned int idx  = idx1+idx2*N;
  
  unsigned int tida  = tid1+(tid2+8)*BLOCK_SIZE1;
  unsigned int idxa  = idx1+((idx2+a*8)%N)*N;

  As[tid]=Ai[idx]; As[tida]=Ai[idxa]; 
  __syncthreads();
  if(tid2%2==0)
  {  Bs[tid]   = h0*As[tid]+h1*As[tid+16]+h2*As[tid+32]+h3*As[tid+48]+h4*As[tid+64]+h5*As[tid+80]
               +h6*As[tid+96]+h7*As[tid+112]+h8*As[tid+128];
     Bs[tid+16]= g0*As[tid]+g1*As[tid+16]+g2*As[tid+32]+g3*As[tid+48]+g4*As[tid+64]+g5*As[tid+80]
               +g6*As[tid+96]; }
  __syncthreads();
  Ad[idx] = Bs[tid];
}

int main()
{  BASETYPE *A, *B, *A1, *Ai, *Ae; int i, j, k, a;
   double error; struct timeval start, end;
   A = (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   B = (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   A1= (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   if((A==NULL)||(B==NULL)||(A1==NULL)) { fprintf(stderr,"No memory\n"); exit(-1); }

   // Input data
   srand48(0); for(i=0;i<N2;i++) A[i]=drand48(); // Example of input data

   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ai,N2*sizeof(BASETYPE));
   cudaMalloc((void**)&Ae,N2*sizeof(BASETYPE));
   cudaMemcpy(Ai,A,N2*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Ae,A,N2*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   dim3 dimGridN0(N2/BLOCK_SIZE);
   dim3 dimBlock0(BLOCK_SIZE);

   // Computation
   gettimeofday(&start,NULL);
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  dim3 dimGridAu(N/(a*BLOCK_SIZE1),N/(a*BLOCK_SIZE1));
      dim3 dimBlock0u(BLOCK_SIZE1,BLOCK_SIZE1);
      dim3 dimBlock1u(8,BLOCK_SIZE1); 
      dim3 dimBlock2u(BLOCK_SIZE1,8); 
      split_rows<<<dimGridAu,dimBlock0u>>>(Ai,Ae,a);
      merge_rows<<<dimGridAu,dimBlock1u>>>(Ai,Ae,a);
      split_columns<<<dimGridAu,dimBlock0u>>>(Ae,Ai,a);
      merge_columns<<<dimGridAu,dimBlock2u>>>(Ae,Ai,a);
   } 
   cudaThreadSynchronize(); 
   gettimeofday(&end,NULL);
   printf("GPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
     
   cudaMemcpy(A1,Ai,N2*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ai); cudaFree(Ae); 
   cuda_error("\n\nOJO\nhubo errores\n\n");
   
   // CPU test
   gettimeofday(&start,NULL);
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  for(j=0;j<N;j+=a)
      {  for(i=0;i<N;i+=2*a) 
         { B[j*N+i]=   h0*A[j*N+i]+h1*A[j*N+i+a]+h2*A[j*N+(i+2*a)%N]+h3*A[j*N+(i+3*a)%N]
	     +h4*A[j*N+(i+4*a)%N]+h5*A[j*N+(i+5*a)%N]+h6*A[j*N+(i+6*a)%N]+h7*A[j*N+(i+7*a)%N]
             +h8*A[j*N+(i+8*a)%N];
           B[j*N+i+a]= g0*A[j*N+i]+g1*A[j*N+i+a]+g2*A[j*N+(i+2*a)%N]+g3*A[j*N+(i+3*a)%N]
	     +g4*A[j*N+(i+4*a)%N]+g5*A[j*N+(i+5*a)%N]+g6*A[j*N+(i+6*a)%N]; }}
      for(j=0;j<N;j+=a)
      {  for(i=0;i<N2;i+=2*a*N)   
         { A[j+i]=     h0*B[j+i]+h1*B[j+i+a*N]+h2*B[(j+i+2*a*N)%N2]+h3*B[(j+i+3*a*N)%N2]
	     +h4*B[(j+i+4*a*N)%N2]+h5*B[(j+i+5*a*N)%N2]+h6*B[(j+i+6*a*N)%N2]+h7*B[(j+i+7*a*N)%N2]
	     +h8*B[(j+i+8*a*N)%N2];
           A[j+i+a*N]= g0*B[j+i]+g1*B[j+i+a*N]+g2*B[(j+i+2*a*N)%N2]+g3*B[(j+i+3*a*N)%N2]
	     +g4*B[(j+i+4*a*N)%N2]+g5*B[(j+i+5*a*N)%N2]+g6*B[(j+i+6*a*N)%N2]; }}}  
   gettimeofday(&end,NULL);
   printf("CPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   error=0; for(i=0;i<N2;i++) error += fabs(A[i]-A1[i]);
 
   printf("Wavelet 97, FBS, blocks 16x16 (SM) and 16x16 (SM)\n");
   printf("Size=%dx%d, Stages=%d, Floatsize=%ld\n",N,N,STAGES,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(CPU-GPU)= %f\n",error);

   free(A); free(B); free(A1);
}
