// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Wavelet: 2-D Cohen-Daubechies-Feauveau 9-7 
// Method: lifting
// Size of matrix: power of two
// Data extension: cyclical 
// Ordering: even and odd coefficients interleaved 
// GPU code: split-and-merge, blocks 256x1 (horizontal) and 16x16 (vertical) 
// Compile: nvcc -arch=sm_20 -O3 wavelet2D_97_split_and_merge.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define N (1<<11)
#define N2 (N*N)        // Size of matrix: power of two
#define STAGES 4        // Number of stages of the wavelet
#define W 0             // Split-and-merge parameter
#define BASETYPE double // float or double
#define BLOCK_SIZE1 16  
#define BLOCK_SIZE (BLOCK_SIZE1*BLOCK_SIZE1)

#define Xa (-1.586134342)
#define Xb (-0.05298011854)
#define Xc ( 0.8829110762)
#define Xd ( 0.4435068522)
#define Xe ( 1.149604398)
#define Xf (1/Xe)

void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void steps_final(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE]; BASETYPE b;
  unsigned int tid = threadIdx.x;
  unsigned int idx = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x);
  Ai+= a*N*blockIdx.y; Ad+= a*N*blockIdx.y;
  
  As[tid]=Ai[idx]; 
  __syncthreads();

  if(tid%2==1) As[tid] += Xa*(As[(tid-1+blockDim.x)%blockDim.x]+As[(tid+1)%blockDim.x]);
  __syncthreads();
  if(tid%2==0) As[tid] += Xb*(As[(tid-1+blockDim.x)%blockDim.x]+As[(tid+1)%blockDim.x]);
  __syncthreads();
  if(tid%2==1) As[tid] += Xc*(As[(tid-1+blockDim.x)%blockDim.x]+As[(tid+1)%blockDim.x]);
  __syncthreads();
  if(tid%2==0) As[tid] += Xd*(As[(tid-1+blockDim.x)%blockDim.x]+As[(tid+1)%blockDim.x]);
  __syncthreads(); 

  if(tid%2==0) b=Xe; else b=Xf;
  As[tid]=b*As[tid];
  __syncthreads(); 

  Ad[idx]= As[tid]; 
}

__global__ void split_rows(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE-W]; BASETYPE b;
  unsigned int tid = 2*threadIdx.x;
  unsigned int tid1 = threadIdx.x;
  unsigned int idx1 = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x);
  Ai+= a*N*blockIdx.y; Ad+= a*N*blockIdx.y;
  
  As[tid1]=Ai[idx1];As[tid1+(BLOCK_SIZE-W)/2]=Ai[idx1+a*(BLOCK_SIZE-W)/2];
  __syncthreads();
  if(tid<BLOCK_SIZE-W-2)            As[tid+1] += Xa*(As[tid]+As[tid+2]);
  __syncthreads();
  if((tid>1)&&(tid<BLOCK_SIZE-W-2)) As[tid]   += Xb*(As[tid-1]+As[tid+1]);
  __syncthreads();
  if((tid>1)&&(tid<BLOCK_SIZE-W-4)) As[tid+1] += Xc*(As[tid]+As[tid+2]);
  __syncthreads();
  if((tid>3)&&(tid<BLOCK_SIZE-W-4)) As[tid]   += Xd*(As[tid-1]+As[tid+1]);
  __syncthreads(); 

  if(tid1%2==0) b=Xe; else b=Xf;
  if(tid1>3) As[tid1]=b*As[tid1];
  if(tid1<(BLOCK_SIZE-W)/2-5) As[tid1+(BLOCK_SIZE-W)/2]=b*As[tid1+(BLOCK_SIZE-W)/2];
  __syncthreads(); 

  Ad[idx1]= As[tid1]; Ad[idx1+a*(BLOCK_SIZE-W)/2]= As[tid1+(BLOCK_SIZE-W)/2];
}

__global__ void merge_rows(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[10+W]; BASETYPE b;
  unsigned int tid = 2*threadIdx.x;
  unsigned int tid1 = threadIdx.x;
  unsigned int idx1 = (a*(BLOCK_SIZE * (blockIdx.x+1) + threadIdx.x-5-W))%N;
  Ai+= a*N*blockIdx.y; Ad+= a*N*blockIdx.y;

  As[tid1]=Ai[idx1];  As[tid1+(10+W)/2]=Ai[(idx1+a*(10+W)/2)%N]; 
  __syncthreads();

  if((tid>3)&&(tid<5+W)) As[tid]   += Xa*(As[tid+1]+As[tid-1]);
  __syncthreads();
  if((tid>1)&&(tid<5+W)) As[tid+1] += Xb*(As[tid+2]+As[tid]);
  __syncthreads();
  if((tid>1)&&(tid<7+W)) As[tid]   += Xc*(As[tid+1]+As[tid-1]);
  __syncthreads();
  if((tid<7+W))          As[tid+1] += Xd*(As[tid+2]+As[tid]);
  __syncthreads();

  if(tid1%2==1) b=Xe; else b=Xf;
  Ad[idx1] = b*As[tid1]; 
  if(tid1%2==0) b=Xe; else b=Xf;
  if(tid1<4+W/2) Ad[(idx1+a*(10+W)/2)%N] = b*As[tid1+(10+W)/2];
 }

__global__ void split_columns(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE1*BLOCK_SIZE1]; BASETYPE b;
  unsigned int tid1 = threadIdx.x;
  unsigned int tid2 = threadIdx.y;
  unsigned int idx1 = a*(threadIdx.x + BLOCK_SIZE1 * blockIdx.x);
  unsigned int idx2 = a*(threadIdx.y + BLOCK_SIZE1 * blockIdx.y);
  unsigned int tid  = tid1+tid2*BLOCK_SIZE1;
  unsigned int idx  = idx1+idx2*N;
  
  As[tid]=Ai[idx];
  __syncthreads();

  if((tid2%2==1)&&(tid2>0)&&(tid2<BLOCK_SIZE1-1)) As[tid] += Xa*(As[tid-16]+As[tid+16]);
  __syncthreads();
  if((tid2%2==0)&&(tid2>1)&&(tid2<BLOCK_SIZE1-2)) As[tid] += Xb*(As[tid-16]+As[tid+16]);
  __syncthreads();
  if((tid2%2==1)&&(tid2>2)&&(tid2<BLOCK_SIZE1-3)) As[tid] += Xc*(As[tid-16]+As[tid+16]);
  __syncthreads();
  if((tid2%2==0)&&(tid2>3)&&(tid2<BLOCK_SIZE1-4)) As[tid] += Xd*(As[tid-16]+As[tid+16]);
  __syncthreads(); 

  if(tid2%2==0) b=Xe; else b=Xf;
  if((tid2>3)&&(tid2<BLOCK_SIZE1-5)) As[tid]=b*As[tid];
  __syncthreads(); 

  Ad[idx]= As[tid];
}

__global__ void merge_columns(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[9*BLOCK_SIZE1]; BASETYPE b;
  unsigned int tid1 = threadIdx.x;
  unsigned int tid2 = threadIdx.y;
  unsigned int idx1 = a*(threadIdx.x + BLOCK_SIZE1 * blockIdx.x);
  unsigned int idx2 = (a*(BLOCK_SIZE1 * (blockIdx.y+1) + threadIdx.y-5))%N;
  unsigned int tid  = tid1+tid2*BLOCK_SIZE1;
  unsigned int idx  = idx1+idx2*N;
  
  As[tid]=Ai[idx]; 
  __syncthreads();

  // original
  if((tid2%2==0)&&(tid2>3)&&(tid2<5)) As[tid] += Xa*(As[tid+16]+As[tid-16]);
  __syncthreads();
  if((tid2%2==1)&&(tid2>2)&&(tid2<6)) As[tid] += Xb*(As[tid+16]+As[tid-16]);
  __syncthreads();
  if((tid2%2==0)&&(tid2>1)&&(tid2<7)) As[tid] += Xc*(As[tid+16]+As[tid-16]);
  __syncthreads();
  if((tid2%2==1)&&(tid2>0)&&(tid2<8)) As[tid] += Xd*(As[tid+16]+As[tid-16]);
  __syncthreads();

  if(tid2%2==1) b=Xe; else b=Xf;
  Ad[idx] = b*As[tid];
 }

int main()
{  BASETYPE *A, *A1, *Ai; int i, j, k, a, b;
   double error; struct timeval start, end;
   A = (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   A1= (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   if((A==NULL)||(A1==NULL)) { fprintf(stderr,"No memory\n"); exit(-1); }

   // Input data
   srand48(0); for(i=0;i<N2;i++) A[i]=drand48(); // Example of input data

   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ai,N2*sizeof(BASETYPE));
   cudaMemcpy(Ai,A,N2*sizeof(BASETYPE),cudaMemcpyHostToDevice);

   dim3 dimGridN0(N2/BLOCK_SIZE);
   dim3 dimBlock0(BLOCK_SIZE);
   dim3 dimBlock0u(BLOCK_SIZE1,BLOCK_SIZE1);
   dim3 dimBlock1u(9,BLOCK_SIZE1);
   dim3 dimBlock2u(BLOCK_SIZE1,9);
   dim3 dimBlock3u(BLOCK_SIZE/2,1);
   dim3 dimBlock4u(10/2,1);

   // Computation
   gettimeofday(&start,NULL);
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  b=N/(a*BLOCK_SIZE); if(b>1)
      {  dim3 dimGridBu(b,N/a);
         split_rows<<<dimGridBu,dimBlock3u>>>(Ai,Ai,a); 
         merge_rows<<<dimGridBu,dimBlock4u>>>(Ai,Ai,a); }
      else { dim3 dimGridBu(1,N/a); dim3 dimBlock33u(N/a,1);
	 steps_final<<<dimGridBu,dimBlock33u>>>(Ai,Ai,a); }
      dim3 dimGridAu(N/(a*BLOCK_SIZE1),N/(a*BLOCK_SIZE1));
      split_columns<<<dimGridAu,dimBlock0u>>>(Ai,Ai,a);
      merge_columns<<<dimGridAu,dimBlock2u>>>(Ai,Ai,a);
   }  
   cudaThreadSynchronize(); 
   gettimeofday(&end,NULL);
   printf("GPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
    
   // Write results
   cudaMemcpy(A1,Ai,N2*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ai); 
   cuda_error("\n\nOJO\nhubo errores\n\n");
   
   // CPU test
   gettimeofday(&start,NULL);
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  for(j=0;j<N;j+=a)
      {  for(i=a;i<N;i+=2*a) A[j*N+i] += Xa*(A[j*N+i-a]+A[j*N+(i+a)%N]);
         for(i=0;i<N;i+=2*a) A[j*N+i] += Xb*(A[j*N+i+a]+A[j*N+(i-a+N)%N]);  
         for(i=a;i<N;i+=2*a) A[j*N+i] += Xc*(A[j*N+i-a]+A[j*N+(i+a)%N]);
         for(i=0;i<N;i+=2*a) A[j*N+i] += Xd*(A[j*N+i+a]+A[j*N+(i-a+N)%N]); 
         for(i=0;i<N/(2*a);i++) 
	 {  A[j*N+2*a*i]= Xe*A[j*N+2*a*i]; A[j*N+2*a*i+a]= Xf*A[j*N+2*a*i+a]; }}
      for(j=0;j<N;j+=a)
      {  for(i=a*N;i<N2;i+=2*a*N) A[j+i] += Xa*(A[j+i-a*N]+A[j+(i+a*N)%N2]);
         for(i=0;i<N2;i+=2*a*N)   A[j+i] += Xb*(A[j+i+a*N]+A[j+(i-a*N+N2)%N2]);  
         for(i=a*N;i<N2;i+=2*a*N) A[j+i] += Xc*(A[j+i-a*N]+A[j+(i+a*N)%N2]);
         for(i=0;i<N2;i+=2*a*N)   A[j+i] += Xd*(A[j+i+a*N]+A[j+(i-a*N+N2)%N2]); 
         for(i=0;i<N2/(2*a*N);i++) 
	 {  A[j+2*a*N*i]= Xe*A[j+2*a*N*i]; A[j+2*a*N*i+a*N]= Xf*A[j+2*a*N*i+a*N]; } } }  
   gettimeofday(&end,NULL);
   printf("CPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   error=0; for(i=0;i<N2;i++) error += fabs(A[i]-A1[i]);

   printf("Wavelet 97, lifting, blocks 256x1 (SM) and 16x16 (SM)\n");
   printf("Size=%dx%d, Stages=%d, Floatsize=%ld\n",N,N,STAGES,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(CPU-GPU)= %f\n",error);

   free(A); free(A1);
}
