// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Wavelet: 2-D Daubechies D4 
// Method: filter-bank
// Size of matrix: power of two
// Data extension: cyclical 
// Ordering: even and odd coefficients interleaved 
// GPU code: horizontal (split-and-merge, blocks 256x1, halt of threads) 
//           vertical (global memory, blocks 256x1, halt of threads)
// Compile: nvcc -arch=sm_20 -O3 wavelet2D_D4FBS_split_and_merge.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define N (1<<11)
#define N2 (N*N)        // Size of matrix: power of two
#define STAGES 4        // Number of stages of the wavelet
#define W 0             // split-and-merge parameter
#define BASETYPE double // float or double
#define BLOCK_SIZE 256

#define h0 ( 0.48296291314453414337)
#define h1 ( 0.83651630373780790557)
#define h2 ( 0.22414386804201338102)
#define h3 (-0.12940952255126038117)

void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void split_rows(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE-W], Bs[BLOCK_SIZE-W];
  unsigned int tid = 2*threadIdx.x;
  unsigned int tid1 = threadIdx.x;
  unsigned int idx1 = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x);
  Ai+= a*N*blockIdx.y; Ad+= a*N*blockIdx.y;
  
  As[tid1]=Ai[idx1]; As[tid1+(BLOCK_SIZE-W)/2]=Ai[idx1+a*(BLOCK_SIZE-W)/2];
  __syncthreads();
  if(tid<BLOCK_SIZE-W-3) 
  {  Bs[tid]   = h0*As[tid]+h1*As[tid+1]+h2*As[tid+2]+h3*As[tid+3];
     Bs[tid+1] = h3*As[tid]-h2*As[tid+1]+h1*As[tid+2]-h0*As[tid+3]; }
  __syncthreads();
  
  Ad[idx1]= Bs[tid1]; Ad[idx1+a*(BLOCK_SIZE-W)/2]= Bs[tid1+(BLOCK_SIZE-W)/2];
}

__global__ void merge_rows(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As[BLOCK_SIZE], Bs[BLOCK_SIZE];
  unsigned int tid1 = threadIdx.x;
  unsigned int tid2 = threadIdx.y;
  unsigned int idx1 = (a*(BLOCK_SIZE * (blockIdx.x+1) + threadIdx.x-2))%N;
  unsigned int idx2 = a*(threadIdx.y + (BLOCK_SIZE/4) * blockIdx.y);
  unsigned int tid  = tid1+tid2*4;
  unsigned int idx  = idx1+idx2*N;
  
  As[tid]=Ai[idx]; 
  __syncthreads();

  if(tid1==0)
  {  Bs[tid]   = h0*As[tid]+h1*As[tid+1]+h2*As[tid+2]+h3*As[tid+3];
     Bs[tid+1] = h3*As[tid]-h2*As[tid+1]+h1*As[tid+2]-h0*As[tid+3]; }
  __syncthreads();
  
  if(tid1<2) Ad[idx] = Bs[tid];
}

__global__ void step_columns(BASETYPE* Ai, BASETYPE* Ad, int a)
{ unsigned int idx = (2*a*blockIdx.y)*N+a*(threadIdx.x+BLOCK_SIZE*blockIdx.x);
  unsigned int idx1 = (idx+a*N)%N2;
  unsigned int idx2 = (idx+2*a*N)%N2;
  unsigned int idx3 = (idx+3*a*N)%N2;
  Ad[idx]  = h0*Ai[idx]+h1*Ai[idx1]+h2*Ai[idx2]+h3*Ai[idx3];
  Ad[idx1] = h3*Ai[idx]-h2*Ai[idx1]+h1*Ai[idx2]-h0*Ai[idx3];
}

int main()
{  BASETYPE *A, *B, *A1, *Ai, *Ae; int i, j, k, a;
   double error; struct timeval start, end;
   A = (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   B = (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   A1= (BASETYPE *) malloc(N2*sizeof(BASETYPE));
   if((A==NULL)||(B==NULL)||(A1==NULL)) { fprintf(stderr,"No memory\n"); exit(-1); }
   
   // Input data
   srand48(0); for(i=0;i<N2;i++) A[i]=drand48(); // Example of input data

   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ai,N2*sizeof(BASETYPE));
   cudaMalloc((void**)&Ae,N2*sizeof(BASETYPE));
   cudaMemcpy(Ai,A,N2*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   cudaMemcpy(Ae,A,N2*sizeof(BASETYPE),cudaMemcpyHostToDevice);

   dim3 dimGridN0(N2/BLOCK_SIZE);
   dim3 dimBlock0(BLOCK_SIZE);
   dim3 dimBlock3u(BLOCK_SIZE/2,1);
   dim3 dimBlock4u(1,1);

   // Computation
   gettimeofday(&start,NULL);
   for(k=0,a=1;k<STAGES;k++,a=2*a)
   {  dim3 dimGridAu(N/(a*BLOCK_SIZE),N/a);
      split_rows<<<dimGridAu,dimBlock3u>>>(Ai,Ae,a);
      dim3 dimGridBu(N/(a*BLOCK_SIZE),N/(a*(BLOCK_SIZE/4)));
      dim3 dimBlock1u(4,BLOCK_SIZE/4);
      merge_rows<<<dimGridBu,dimBlock1u>>>(Ai,Ae,a);
      dim3 dimGrid0u(N/(a*BLOCK_SIZE),N/(2*a));
      dim3 dimBlock0u(BLOCK_SIZE,1);
      step_columns<<<dimGrid0u,dimBlock0u>>>(Ae,Ai,a);  
   }   
   cudaThreadSynchronize(); 
   gettimeofday(&end,NULL);
   printf("GPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
     
   cudaMemcpy(A1,Ai,N2*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ai); cudaFree(Ae);
   cuda_error("\n\nOJO\nhubo errores\n\n");
   
   // CPU test
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  for(j=0;j<N;j+=a)
      {  for(i=0;i<N;i+=2*a) 
         { B[j*N+i]=   h0*A[j*N+i]+h1*A[j*N+i+a]+h2*A[j*N+(i+2*a)%N]+h3*A[j*N+(i+3*a)%N];
           B[j*N+i+a]= h3*A[j*N+i]-h2*A[j*N+i+a]+h1*A[j*N+(i+2*a)%N]-h0*A[j*N+(i+3*a)%N]; }}
      for(j=0;j<N;j+=a)
      {  for(i=0;i<N2;i+=2*a*N)   
         { A[j+i]=     h0*B[j+i]+h1*B[j+i+a*N]+h2*B[(j+i+2*a*N)%N2]+h3*B[(j+i+3*a*N)%N2];
           A[j+i+a*N]= h3*B[j+i]-h2*B[j+i+a*N]+h1*B[(j+i+2*a*N)%N2]-h0*B[(j+i+3*a*N)%N2]; }}}  
   gettimeofday(&end,NULL);
   printf("CPU t=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   error=0; for(i=0;i<N2;i++) error += fabs(A[i]-A1[i]);

   printf("Wavelet D4 LBS, blocks 256x1 (SM) and 256x1 (global memory), half of threads\n"); 
   printf("Size=%dx%d, Stages=%d. Floatsize=%ld\n",N,N,STAGES,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(CPU-GPU)= %f\n",error);

   free(A); free(B); free(A1);
}
