// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Wavelet: Cohen-Daubechies-Feauveau 9-7 
// Method: lifting
// Size of sequence: power of two
// Data extension: cyclical 
// Ordering: even and odd coefficients interleaved 
// CPU code
// Compile: gcc -O3 wavelet97.c -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define N (1<<20)      // Size of the sequence (power of 2)
#define STAGES 8       // Number of stages of the wavelet
#define PRINT 0        // 1=print data 0=do not print data
#define BASETYPE float // float or double

#define Xa -1.586134342
#define Xb -0.05298011854
#define Xc 0.8829110762
#define Xd 0.4435068522
#define Xe 1.149604398
#define Xf (1/Xe)

int main()
{  BASETYPE *A, *Ain; int i, k, a; 
   BASETYPE error; struct timeval start, end;
   A = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   Ain = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   if((A==NULL)||(Ain==NULL)) { fprintf(stderr,"No memory\n"); exit(-1); }

   // Input data
   srand48(0); for(i=0;i<N;i++) Ain[i]=drand48(); // Example of input data
   memcpy(A,Ain,N*sizeof(BASETYPE));
   if(PRINT)
   {  printf("\nOriginal signal:\n");
      for(i=0;i<N;i++) printf("%3.2f, ",A[i]); printf("\n"); }
  
   // Wavelet
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  for(i=a;i<N;i+=2*a) A[i] += Xa*(A[i-a]+A[(i+a)%N]);
      for(i=0;i<N;i+=2*a) A[i] += Xb*(A[i+a]+A[(i-a+N)%N]);  
      for(i=a;i<N;i+=2*a) A[i] += Xc*(A[i-a]+A[(i+a)%N]);
      for(i=0;i<N;i+=2*a) A[i] += Xd*(A[i+a]+A[(i-a+N)%N]);  
      for(i=0;i<N/(2*a);i++) { A[2*a*i]= Xe*A[2*a*i]; A[2*a*i+a]= Xf*A[2*a*i+a]; }
   } 
   if(PRINT)
   {  printf("\nWavelet 9-7:\n"); 
      for(i=0;i<N;i++) printf("%3.2f, ",A[i]); printf("\n"); }

   // Inverse wavelet
   for(k=0,a=1<<(STAGES-1);k<STAGES;k++,a=a/2)
   {  for(i=0;i<N/(2*a);i++) { A[2*a*i]= Xf*A[2*a*i]; A[2*a*i+a]= Xe*A[2*a*i+a]; }
      for(i=0;i<N;i+=2*a) A[i] += -Xd*(A[i+a]+A[(i-a+N)%N]);  
      for(i=a;i<N;i+=2*a) A[i] += -Xc*(A[i-a]+A[(i+a)%N]);
      for(i=0;i<N;i+=2*a) A[i] += -Xb*(A[i+a]+A[(i-a+N)%N]);  
      for(i=a;i<N;i+=2*a) A[i] += -Xa*(A[i-a]+A[(i+a)%N]);
   } 

   // Output
   if(PRINT)
   {  printf("\nInverse Wavelet 9-7:\n"); 
      for(i=0;i<N;i++) printf("%3.2f, ",A[i]); printf("\n");  }
   error=0; for(i=0;i<N;i++) error += fabs(A[i]-Ain[i]);

   printf("Wavelet 9-7, lifting\n"); 
   printf("Size=%d, Stages=%d\n",N,STAGES);
   printf("Acumulative error(ORI-INV)= %f\n",error);

   free(A); free(Ain);
   return(0);
}
