// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Wavelet: Cohen-Daubechies-Feauveau 9-7 
// Method: lifting
// Size of sequence: power of two
// Data extension: cyclical 
// Ordering: even and odd coefficients interleaved 
// GPU code: global memory
// Compile: nvcc -arch=sm_20 -O3 wavelet97_global_memory.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define N (1<<20)       // Size of the sequence: power of two
#define STAGES 8        // Number of stages of the wavelet
#define BLOCK_SIZE 256  // Size of the CUDA BLOCK 
#define W 0             // Split-and-merge parameter
#define BASETYPE double  // float or double

#define Xa -1.586134342
#define Xb -0.05298011854
#define Xc 0.8829110762
#define Xd 0.4435068522
#define Xe 1.149604398
#define Xf (1/Xe)

void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void wavelet_scale(BASETYPE* Ai, BASETYPE* Ad, int a)
{ unsigned int tid=threadIdx.x;
  unsigned int idx = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x);
  BASETYPE b;
  if(tid%2==0) b=Xe; else b=Xf;
  Ad[idx] = b*Ai[idx];
}

__global__ void wavelet_step(BASETYPE* Ai, BASETYPE* Ad, int a, int s, BASETYPE XX)
{ unsigned int tid=threadIdx.x;
  unsigned int idx = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x);
  unsigned int idx1 = (idx+a)%N;
  unsigned int idx2 = (idx+N-a)%N;
  if(tid%2==s) Ad[idx] += XX*(Ai[idx1]+Ai[idx2]);
}

main()
{  BASETYPE *A, *A1, *Ai; int i, k, a;
   double error; struct timeval start, end;
   A = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   A1= (BASETYPE *) malloc(N*sizeof(BASETYPE));
   if((A==NULL)||(A1==NULL)) { fprintf(stderr,"No memory\n"); exit(-1); }
   
   // Input data
   srand48(0); for(i=0;i<N;i++) A[i]=drand48(); // Example of input data
   
   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ai,N*sizeof(BASETYPE));
   cudaMemcpy(Ai,A,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   dim3 dimBlock0(BLOCK_SIZE);
   dim3 dimGrid0(N/BLOCK_SIZE);

   // Computation
   gettimeofday(&start,NULL);
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  dim3 dimGridA(N/(a*BLOCK_SIZE));
      wavelet_step<<<dimGridA,dimBlock0>>>(Ai,Ai,a,1,Xa); 
      wavelet_step<<<dimGridA,dimBlock0>>>(Ai,Ai,a,0,Xb); 
      wavelet_step<<<dimGridA,dimBlock0>>>(Ai,Ai,a,1,Xc); 
      wavelet_step<<<dimGridA,dimBlock0>>>(Ai,Ai,a,0,Xd); 
      wavelet_scale<<<dimGridA,dimBlock0>>>(Ai,Ai,a);  
   }
   cudaThreadSynchronize(); 
   gettimeofday(&end,NULL);
     
   // Write results
   cudaMemcpy(A1,Ai,N*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ai); 
   cuda_error("\n\nExecution error\n\n");
   
   // CPU test
   for(k=0,a=1;k<STAGES;k++,a=a*2)
   {  for(i=a;i<N;i+=2*a) A[i] += Xa*(A[i-a]+A[(i+a)%N]);
      for(i=0;i<N;i+=2*a) A[i] += Xb*(A[i+a]+A[(i-a+N)%N]);  
      for(i=a;i<N;i+=2*a) A[i] += Xc*(A[i-a]+A[(i+a)%N]);
      for(i=0;i<N;i+=2*a) A[i] += Xd*(A[i+a]+A[(i-a+N)%N]);  
      for(i=0;i<N/(2*a);i++) {A[2*a*i]= Xe*A[2*a*i]; A[2*a*i+a]= Xf*A[2*a*i+a]; }
   } 
   error=0; for(i=0;i<N;i++) error += fabs(A[i]-A1[i]);

   printf("Wavelet 9-7, lifting, global memory\n"); 
   printf("Size=%d, Stages=%d, Floatsize=%ld\n",N,STAGES,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(GPU-CPU)= %f\n",error);
   printf("Execution time=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   
   free(A); free(A1);
}
