// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Wavelet: Daubechies D4 
// Method: filter bank
// Size of sequence: power of two
// Data extension: cyclical 
// Ordering: even and odd coefficients interleaved 
// CPU code
// Compile: gcc -O3 waveletD4_FB.c -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define N (1<<20)      // Size of the sequence (power of 2)
#define STAGES 8       // Number of stages of the wavelet
#define PRINT 0        // 1=print data 0=no imprime data
#define BASETYPE float // float or double

#define h0 ((1+sqrt(3))/(4*sqrt(2)))
#define h1 ((3+sqrt(3))/(4*sqrt(2)))
#define h2 ((3-sqrt(3))/(4*sqrt(2)))
#define h3 ((1-sqrt(3))/(4*sqrt(2)))

int main()
{  BASETYPE *A, *B,  *Ain; int i, j, k, a;
   double error;  struct timeval start, end;
   A = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   B = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   Ain= (BASETYPE *) malloc(N*sizeof(BASETYPE));
   if((A==NULL)||(B==NULL)||(Ain==NULL)) { fprintf(stderr,"No memory\n"); exit(-1); }

   // Input data
   srand48(0); for(i=0;i<N;i++) Ain[i]=drand48(); // Example of input data
   memcpy(A,Ain,N*sizeof(BASETYPE));
   if(PRINT)
   {  printf("Input:\n");
      for(j=0;j<N;j++) printf("%3.2f, ",A[j]); printf("\n"); }
   
   // Wavelet
   for(k=0,a=1;k<STAGES;k+=2,a=a*4)
   {  for(i=0;i<N;i+=2*a) 
      {  B[i]=   h0*A[i]+h1*A[(i+a)%N]+h2*A[(i+2*a)%N]+h3*A[(i+3*a)%N];
         B[i+a]= h3*A[i]-h2*A[(i+a)%N]+h1*A[(i+2*a)%N]-h0*A[(i+3*a)%N]; }
      for(i=0;i<N;i+=4*a) 
      {  A[i]=     h0*B[i]+h1*B[(i+2*a)%N]+h2*B[(i+4*a)%N]+h3*B[(i+6*a)%N];
         A[i+2*a]= h3*B[i]-h2*B[(i+2*a)%N]+h1*B[(i+4*a)%N]-h0*B[(i+6*a)%N]; 
	 A[i+a]=B[i+a]; A[i+3*a]=B[i+3*a]; } 
   } 
   if(PRINT)
   {  printf("Transform:\n");
      for(j=0;j<N;j++) printf("%3.2f, ",B[j]); printf("\n"); }

   // Inverse wavelet
   for(k=0,a=1<<(STAGES-2);k<STAGES;k+=2,a=a/4)
   { for(i=0;i<N;i+=4*a) 
      {  B[(i+4*a)%N]=     h2*A[i]+h1*A[(i+2*a)%N]+h0*A[(i+4*a)%N]+h3*A[(i+6*a)%N];
         B[(i+2*a+4*a)%N]= h3*A[i]-h0*A[(i+2*a)%N]+h1*A[(i+4*a)%N]-h2*A[(i+6*a)%N]; 
	 B[i+a]=A[i+a]; B[i+3*a]=A[i+3*a]; } 
   for(i=0;i<N;i+=2*a) 
      {  A[(i+2*a)%N]=   h2*B[i]+h1*B[(i+a)%N]+h0*B[(i+2*a)%N]+h3*B[(i+3*a)%N];
         A[(i+a+2*a)%N]= h3*B[i]-h0*B[(i+a)%N]+h1*B[(i+2*a)%N]-h2*B[(i+3*a)%N]; }
   } 

   // Output
   if(PRINT)
   {   printf("Inverse transform:\n");
       for(j=0;j<N;j++) printf("%3.2f, ",A[j]); printf("\n"); }
   error=0; for(i=0;i<N;i++) error += fabs(A[i]-Ain[i]);

   printf("Wavelet D4, filter bank\n");
   printf("Size=%d, Stages=%d\n",N,STAGES);
   printf("Acumulative error(ORI-INV)= %f\n",error);
   
   free(A); free(B); free(Ain);
}
