// Code related to the paper:
// The split-and-merge method in general purpose computation on GPUs
// F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
// Parallel Computing, Volume 38, Issues 6-7, June-July 2012, Pages 277-288, 
// DOI: 10.1016/j.parco.2012.03.003
//
// Wavelet: Daubechies D4
// Method: filter bank
// Size of sequence: power of two
// Data extension: cyclical 
// Ordering: even and odd coefficients interleaved 
// GPU code: split shared, merge global
// Compile: nvcc -arch=sm_20 -O3 waveletD4_FB_split_and_merge.cu -lm
// Execute: ./a.out

/* (C) 2012-2014 F. Arguello, D.B. Heras, M. Boo, and J. Lamas-Rodriguez 
 *     CITIUS - Dept. of Electronics and Computing
 *     University of Santiago de Compostela
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>

#define N (1<<20)       // Size of the sequence: power of two
#define STAGES 8        // Number of stages of the wavelet
#define BLOCK_SIZE 256  // Size of the CUDA BLOCK 
#define BASETYPE double  // float or double

#define h0  0.48296291314453414337
#define h1  0.83651630373780790557
#define h2  0.22414386804201338102
#define h3 -0.12940952255126038117

void cuda_init_2_0()
{  int num_devices, device, found=0;
   int max_multiprocessors = 0, max_device = 0;
   cudaGetDeviceCount(&num_devices);
   for (device = 0; device < num_devices; device++) {
              cudaDeviceProp properties;
              cudaGetDeviceProperties(&properties, device);
 	      printf("Card %d: %s,",device,properties.name);
	      printf(" Capability=%d.%d, ",properties.major,properties.minor);
              if (max_multiprocessors < properties.multiProcessorCount
	           && properties.major >= 2 ) {
		      found=1;
                      max_multiprocessors = properties.multiProcessorCount;
                      max_device = device; } }
   if(found) { cudaSetDevice(max_device); 
               printf("Using card num. %d\n",max_device); }
   else { fprintf(stderr,"No cards with capability >= 2.0 found\n"); exit(0); }
}

void cuda_error(const char *msg)
{   cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {   fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(EXIT_FAILURE); } }
// ---------------------------------------------------------------------  //

__global__ void merge_global(BASETYPE* Ai, BASETYPE* Ad, int a)
{ unsigned int tid=threadIdx.x;
  unsigned int idx = (a*(threadIdx.x+BLOCK_SIZE*(blockIdx.x+1)-8))%N;
  unsigned int idx1 = (idx+a)%N;
  unsigned int idx2 = (idx+2*a)%N;
  unsigned int idx3 = (idx+3*a)%N;

  // one stage
  if(tid==6) {  Ad[idx]  = h0*Ai[idx]+h1*Ai[idx1]+h2*Ai[idx2]+h3*Ai[idx3]; 
                Ad[idx1] = h3*Ai[idx]-h2*Ai[idx1]+h1*Ai[idx2]-h0*Ai[idx3]; 
		Ai[idx]=Ad[idx]; Ai[idx1]=Ad[idx1]; }
  __syncthreads(); 
  
  // two stages
  idx1 = (idx+2*a)%N;  idx2 = (idx+4*a)%N;  idx3 = (idx+6*a)%N;
  if(tid%4==0) { Ai[idx]  = h0*Ad[idx]+h1*Ad[idx1]+h2*Ad[idx2]+h3*Ad[idx3];
                 Ai[idx1] = h3*Ad[idx]-h2*Ad[idx1]+h1*Ad[idx2]-h0*Ad[idx3]; }		  
  else if(tid==9) { Ai[idx]= Ad[idx]; Ai[idx1]= Ad[idx1]; }
} 

__global__ void split_shared(BASETYPE* Ai, BASETYPE* Ad, int a)
{ __shared__ BASETYPE As1[BLOCK_SIZE], As2[BLOCK_SIZE];
  unsigned int tid=threadIdx.x;
  unsigned int idx = a*(threadIdx.x + BLOCK_SIZE * blockIdx.x);
  unsigned int tid1 = tid+1;
  unsigned int tid2 = tid+2;
  unsigned int tid3 = tid+3;

  As1[tid]=Ai[idx];
  __syncthreads(); 
  
  // one stage
  if((tid%2==0)&&(tid<BLOCK_SIZE-2)) 
  {  As2[tid]  = h0*As1[tid]+h1*As1[tid1]+h2*As1[tid2]+h3*As1[tid3];
     As2[tid1] = h3*As1[tid]-h2*As1[tid1]+h1*As1[tid2]-h0*As1[tid3]; }
  __syncthreads(); 
  if((tid%2==0)&&(tid!=0)&&(tid<BLOCK_SIZE-2)) As1[tid1]=As2[tid1];   
  __syncthreads(); 
  
  // two stage
  tid1 = tid+2; tid2 = tid+4; tid3 = tid+6;
  if((tid%4==0)&&(tid>2)&&(tid<BLOCK_SIZE-10)) 
  {  As1[tid]  = h0*As2[tid]+h1*As2[tid1]+h2*As2[tid2]+h3*As2[tid3];
     As1[tid1] = h3*As2[tid]-h2*As2[tid1]+h1*As2[tid2]-h0*As2[tid3]; }
  __syncthreads(); 
  
  if((tid<=6)||(tid>=BLOCK_SIZE-8)) Ad[idx]=As2[tid];  
  Ai[idx]=As1[tid];
}

int main()
{  BASETYPE *A, *B,  *A1, *Ai, *Ae; int i, k, a;
   double error;  struct timeval start, end;
   A = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   B = (BASETYPE *) malloc(N*sizeof(BASETYPE));
   A1= (BASETYPE *) malloc(N*sizeof(BASETYPE));
   if((A==NULL)||(B==NULL)||(A1==NULL)) { fprintf(stderr,"No memory\n"); exit(-1); }

   // Input data
   srand48(0); for(i=0;i<N;i++) A[i]=drand48(); // Example of input data
   
   // Load data
   cuda_init_2_0();
   cudaMalloc((void**)&Ai,N*sizeof(BASETYPE));
   cudaMalloc((void**)&Ae,N*sizeof(BASETYPE));
   cudaMemcpy(Ai,A,N*sizeof(BASETYPE),cudaMemcpyHostToDevice);
   dim3 dimBlock0(BLOCK_SIZE); 
   dim3 dimGrid0(N/BLOCK_SIZE); // block for split
   dim3 dimBlock1(10);          // block for merge

   // Computation
   gettimeofday(&start,NULL);
   for(k=0,a=1;k<STAGES;k+=2,a=a*4)
   {  dim3 dimGridA(N/(a*BLOCK_SIZE));
      split_shared<<<dimGridA,dimBlock0>>>(Ai,Ae,a);
      merge_global<<<dimGridA,dimBlock1>>>(Ai,Ae,a);
   } 
   cudaThreadSynchronize(); 
   gettimeofday(&end,NULL);
     
   // Write results
   cudaMemcpy(A1,Ai,N*sizeof(BASETYPE),cudaMemcpyDeviceToHost);
   cudaFree(Ai); cudaFree(Ae); 
   cuda_error("\n\nCuda Error\n\n");
   
   // CPU test
   for(k=0,a=1;k<STAGES;k+=2,a=a*4)
   {  for(i=0;i<N;i+=2*a) 
      {  B[i]=   h0*A[i]+h1*A[(i+a)%N]+h2*A[(i+2*a)%N]+h3*A[(i+3*a)%N];
         B[i+a]= h3*A[i]-h2*A[(i+a)%N]+h1*A[(i+2*a)%N]-h0*A[(i+3*a)%N]; }
      for(i=0;i<N;i+=4*a) 
      {  A[i]=     h0*B[i]+h1*B[(i+2*a)%N]+h2*B[(i+4*a)%N]+h3*B[(i+6*a)%N];
         A[i+2*a]= h3*B[i]-h2*B[(i+2*a)%N]+h1*B[(i+4*a)%N]-h0*B[(i+6*a)%N]; 
	 A[i+a]=B[i+a]; A[i+3*a]=B[i+3*a]; }
   } 
   error=0; for(i=0;i<N;i++) error += abs(A[i]-A1[i]);

   printf("Wavelet D4, filter bank, split shared, merge global\n");
   printf("Size=%d, Stages=%d, Floatsize=%ld\n",N,STAGES,sizeof(BASETYPE));
   printf("Execution without exceptions in GPU\n");
   printf("Acumulative error(CPU-GPU)= %f\n",error);
   printf("Execution time=%f\n",(end.tv_sec-start.tv_sec)+(end.tv_usec-start.tv_usec)*1E-6);
   
   free(A); free(B); free(A1);
   
}
